__all__ = ('InteractionType',)

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase

from ..interaction_metadata import (
    InteractionMetadataApplicationCommand, InteractionMetadataApplicationCommandAutocomplete, InteractionMetadataBase,
    InteractionMetadataFormSubmit, InteractionMetadataMessageComponent
)


@export
class InteractionType(PreinstancedBase, value_type = int):
    """
    The type of an interaction.
    
    Attributes
    ----------
    metadata_type : `type<InteractionMetadataBase>`
        The interaction's respective metadata type.
    
    name : `str`
        The name of the interaction type.
    
    value : `int`
        The identifier value the interaction type.
    
    Type Attributes
    ---------------
    Every predefined interaction type can be accessed as type attribute as well:
    
    +-----------------------------------+-----------------------------------+-------+---------------------------------------------------+
    | Type attribute name               | Name                              | Value | Metadata type                                     |
    +===================================+===================================+=======+===================================================+
    | none                              | none                              | 0     | InteractionMetadataBase                           |
    +-----------------------------------+-----------------------------------+-------+---------------------------------------------------+
    | ping                              | ping                              | 1     | InteractionMetadataBase                           |
    +-----------------------------------+-----------------------------------+-------+---------------------------------------------------+
    | application_command               | application_command               | 2     | InteractionMetadataApplicationCommand             |
    +-----------------------------------+-----------------------------------+-------+---------------------------------------------------+
    | message_component                 | message_component                 | 3     | InteractionMetadataMessageComponent               |
    +-----------------------------------+-----------------------------------+-------+---------------------------------------------------+
    | application_command_autocomplete  | application_command_autocomplete  | 4     | InteractionMetadataApplicationCommandAutocomplete |
    +-----------------------------------+-----------------------------------+-------+---------------------------------------------------+
    | form_submit                       | form_submit                       | 5     | InteractionMetadataFormSubmit                     |
    +-----------------------------------+-----------------------------------+-------+---------------------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    def __new__(cls, value, name = None, metadata_type = None):
        """
        Creates an interaction type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the interaction type.
        
        name : `None | str` = `None`
            The default name of the interaction type.
        
        metadata_type : `None | type<InteractionMetadataBase>`
            The interaction type's respective metadata type.
        """
        if metadata_type is None:
            metadata_type = InteractionMetadataBase
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.metadata_type = metadata_type
        return self
    
    
    none = P(0, 'none', InteractionMetadataBase)
    ping = P(1, 'ping', InteractionMetadataBase)
    application_command = P(2, 'application_command', InteractionMetadataApplicationCommand)
    message_component = P(3, 'message_component', InteractionMetadataMessageComponent)
    application_command_autocomplete = P(
        4, 'application_command_autocomplete', InteractionMetadataApplicationCommandAutocomplete
    )
    form_submit = P(5, 'form_submit', InteractionMetadataFormSubmit)



__all__ = ('get_plugins_like',)

from re import I as re_ignore_case, compile as re_compile, escape as re_escape

from ..constants import PLUGINS


def _plugin_match_sort_key(item):
    """
    Sort key used inside of ``get_plugins_like`` to sort plugins based on their match rate.
    
    Parameters
    ----------
    item : `tuple` (``Plugin``, `tuple` (`int`, `int`))
        The plugin and it's match rate.
    
    Returns
    -------
    match_rate : `tuple` (`int`, `int`)
    """
    return item[1]


def get_plugins_like(name):
    """
    Returns the plugins witch match the given name the most.
    
    Parameters
    ----------
    name : `str`
        An plugin's name.
    
    Returns
    -------
    plugins : `list` of ``Plugin``.
        The matched plugins.
    """
    name = name.replace('-', '_').replace(' ', '_')
    
    pattern = re_compile(re_escape(name), re_ignore_case)
    
    to_sort = []
    
    for plugin in PLUGINS.values():
        plugin_name = plugin.name
        parsed = pattern.search(plugin_name)
        if parsed is None:
            continue
        
        match_start = parsed.start()
        to_sort.append((plugin, (match_start, len(plugin_name))))
    
    if not to_sort:
        return to_sort
    
    to_sort.sort(key = _plugin_match_sort_key)
    return [item[0] for item in to_sort]



__all__ = ()

from scarletio import RichAttributeErrorBaseType


class EncryptionAdapterBase(RichAttributeErrorBaseType):
    """
    Base encryption adapter.
    
    Attributes
    ----------
    key : `bytes`
        Secret key used for encrypting / decrypting.
    """
    __slots__ = ('key', )
    
    available = False
    key_length = 0
    name = ''
    nonce_length = 0
    priority = 0
    
    
    def __new__(cls, key):
        """
        Instantiates a new encryption adapter.
        
        Parameters
        ----------
        key : `bytes`
            Key to use.
        """
        self = object.__new__(cls)
        self.key = key
        return self
    
    
    def __repr__(self):
        """Returns the encryption adapter's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' key = ')
        repr_parts.append(repr(self.key))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def create_send_packet(self, voice_client, plain_data):
        """
        Creates a send packet.
        
        Parameters
        ----------
        voice_client : ``VoiceClient``
            Voice client to send packet with.
        
        pain_data : `bytes-like`
            Data to encrypt.
        
        Returns
        -------
        packet : `bytes`
        """
        return b''
    
    
    def process_received_payload(self, rtp_packet):
        """
        Creates a send packet.
        
        Parameters
        ----------
        voice_client : ``VoiceClient``
            Voice client to send packet with.
        
        rtp_packet : `RTPPacket`
            Data to decrypt.
        
        Returns
        -------
        packet : `bytes`
        """
        return b''



__all__ = ('CONVERSION_ALLOWED_MENTIONS',)

from ....allowed_mentions import AllowedMentionProxy, is_allowed_mentions_valid, parse_allowed_mentions
from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion
from ....role import Role
from ....user import UserBase


class CONVERSION_ALLOWED_MENTIONS(Conversion):
    # Generic
    
    name = 'allowed_mentions'
    name_aliases = None
    expected_types_messages = (
        f'`None`,  ``{AllowedMentionProxy.__name__}``, `str`, ``{UserBase.__name__}``, ``{Role.__name__}``, '
        f'(`list`, `tuple`, `set`) of (`str`, `{UserBase.__name__}`, `{Role.__name__}`)'
    )
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    def set_merger(old_value, new_value):
        return AllowedMentionProxy._create_from_various(old_value) | new_value
    
    set_type = AllowedMentionProxy
    set_type_processor = None
    set_identifier = None
    set_listing_identifier = None
    
    
    def set_validator(value):
        # use helper
        if is_allowed_mentions_valid(value):
            yield value
            return
        
        # No other cases
        return
    
    
    # Reading
    
    get_default = None
    get_processor = None
    
    
    # Serialization
    
    serializer_key = 'allowed_mentions'
    
    
    def serializer_optional(value):
        yield parse_allowed_mentions(value)
    
    
    def serializer_required(value):
        return parse_allowed_mentions(value)
    
    
    # Sorting
    
    sort_priority = 1500



__all__ = ()

from ...field_parsers import bool_parser_factory, entity_id_parser_factory, force_string_parser_factory
from ...field_putters import entity_id_putter_factory, force_bool_putter_factory, force_string_putter_factory
from ...field_validators import bool_validator_factory, entity_id_validator_factory, force_string_validator_factory

from .constants import TAG_LENGTH_MAX, TAG_LENGTH_MIN

# enabled

parse_enabled = bool_parser_factory('identity_enabled', True)
put_enabled = force_bool_putter_factory('identity_enabled')
validate_enabled = bool_validator_factory('enabled', True)

# guild_id

parse_guild_id = entity_id_parser_factory('identity_guild_id')
put_guild_id = entity_id_putter_factory('identity_guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# tag

parse_tag = force_string_parser_factory('tag')
put_tag = force_string_putter_factory('tag')
validate_tag = force_string_validator_factory('tag', TAG_LENGTH_MIN, TAG_LENGTH_MAX)



from .fields import *
from .flags import *
from .invite import *
from .preinstanced import *
from .utils import *


__all__ = (
    *fields.__all__,
    *flags.__all__,
    *invite.__all__,
    *preinstanced.__all__,
    *utils.__all__,
)



__all__ = ('AvatarDecoration', )

from ...bases import IconSlot, IconType, Slotted

from .fields import parse_sku_id, put_sku_id, validate_sku_id


AVATAR_DECORATION_ASSET = IconSlot('asset', 'asset', None, None)


class AvatarDecoration(metaclass = Slotted):
    """
    Represents a user's profile at a guild.
    
    Attributes
    ----------
    asset_hash : `int`
        The respective asset's hash in `uint128`.
    
    asset_type : ``IconType``
        The respective asset's type.
    
    sku_id : `int`
        The stock keeping unit the avatar decoration is part of.
    """
    __slots__ = ('sku_id',)
    
    asset = AVATAR_DECORATION_ASSET
    
    def __new__(
        cls,
        *,
        asset = ...,
        sku_id = ...,
    ):
        """
        Creates a new avatar decoration instance from the given parameters.
        
        Parameters
        ----------
        asset : `None`, ``Icon``, `str`, Optional (Keyword only)
            The channel's asset.
        
        sku_id : `int`, `None`, ``SKU``, Optional (Keyword only)
            The stock keeping unit the avatar decoration is part of.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # asset
        if asset is ...:
            asset = None
        else:
            asset = cls.asset.validate_icon(asset, allow_data = True)
        
        # sku_id
        if sku_id is ...:
            sku_id = 0
        else:
            sku_id = validate_sku_id(sku_id)
        
        # Construct
        self = object.__new__(cls)
        self.asset = asset
        self.sku_id = sku_id
        return self
    
    
    def __repr__(self):
        """Returns the representation of the avatar decoration."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' asset = ')
        repr_parts.append(repr(self.asset))
        
        sku_id = self.sku_id
        if sku_id:
            repr_parts.append(', sku_id = ')
            repr_parts.append(repr(sku_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the avatar decoration's hash value."""
        hash_value = 0
        
        # asset
        hash_value ^= hash(self.asset)
        
        # sku_id
        hash_value ^= self.sku_id
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two avatar decorations are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # asset
        if self.asset != other.asset:
            return False
        
        # sku_id
        if self.sku_id != other.sku_id:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a avatar decoration from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received avatar decoration data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._set_asset(data)
        self.sku_id = parse_sku_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the avatar decoration to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        type(self).asset.put_into(self.asset, data, defaults)
        put_sku_id(self.sku_id, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the avatar decoration.
        
        Returns
        -------
        new : `instance<type<self>>
        """
        new = object.__new__(type(self))
        new.asset_hash = self.asset_hash
        new.asset_type = self.asset_type
        new.sku_id = self.sku_id
        return new
    
    
    def copy_with(
        self, 
        *,
        asset = ...,
        sku_id = ...,
    ):
        """
        Copies the avatar decoration and modifies the defined the defined fields of it.
        
        Parameters
        ----------
        asset : `None`, ``Icon``, `str`, Optional (Keyword only)
            The channel's asset.
        
        sku_id : `int`, `None`, ``SKU``, Optional (Keyword only)
            The stock keeping unit the avatar decoration is part of.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # asset
        if asset is ...:
            asset = self.asset
        else:
            asset = type(self).asset.validate_icon(asset, allow_data = True)
        
        # sku_id
        if sku_id is ...:
            sku_id = self.sku_id
        else:
            sku_id = validate_sku_id(sku_id)
        
        # Construct
        new = object.__new__(type(self))
        new.asset = asset
        new.sku_id = sku_id
        return new
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates a new avatar decoration with it's default attributes set.
        
        Returns
        -------
        self `instance<cls>`
        """
        self = object.__new__(cls)
        self.asset_hash = 0
        self.asset_type = IconType.none
        self.sku_id = 0
        return self



__all__ = ('GuildWidgetChannel',)

from ...bases import DiscordEntity

from .fields import (
    parse_id, parse_name, parse_position, put_id, put_name, put_position, validate_id, validate_name,
    validate_position
)


class GuildWidgetChannel(DiscordEntity):
    """
    Represents a ``GuildWidget``'s channel.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the guild widget channel.
    name : `str`
        The channel's name.
    position : `int`
        The channel's position.
    """
    __slots__ = ('name', 'position')
    
    
    def __new__(
        cls,
        *,
        channel_id = ...,
        name = ...,
        position = ...,
    ):
        """
        Creates a new guild widget channel with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The unique identifier number of the guild widget channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_id(channel_id)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # position
        if position is ...:
            position = 0
        else:
            position = validate_position(position)
        
        # Construct
        self = object.__new__(cls)
        self.id = channel_id
        self.name = name
        self.position = position
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild widget channel from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Guild widget channel data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.position = parse_position(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the guild widget channel.
        
        Parameters
        ----------
        defaults : `bool`
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_id(self.id, data, defaults)
        put_name(self.name, data, defaults)
        put_position(self.position, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the guild widget channel's representation."""
        return f'<{self.__class__.__name__} id = {self.id}, name = {self.name!r}>'
    
    
    def __hash__(self):
        """Returns the guild widget channel's hash value."""
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # position
        hash_value ^= hash(self.position)
        
        return hash_value
    
    
    def __gt__(self, other):
        """
        Whether this guild widget channel has greater (visible) position than the other at their respective guild.
        """
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_greater_than_same_type(other)
    
    
    def __ge__(self, other):
        """
        Whether this guild widget channel has greater or equal (visible) position than the other at their respective
        guild.
        """
        if type(self) is not type(other):
            return NotImplemented
    
        return self._is_greater_than_same_type(other) or self._is_equal_same_type(other)
    
    
    def __eq__(self, other):
        """Returns whether the two guild widget channels are equal."""
        if type(self) is not type(other):
            return NotImplemented
    
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two guild widget channels are not equal."""
        if type(self) is not type(other):
            return NotImplemented
    
        return not self._is_equal_same_type(other)
    
    
    def __le__(self, other):
        """
        Whether this guild widget channel has lower or equal (visible) position than the other at their respective
        guild.
        """
        if type(self) is not type(other):
            return NotImplemented
        
        return other._is_greater_than_same_type(self) or self._is_equal_same_type(other)
    
    
    def __lt__(self, other):
        """
        Whether this guild widget channel has lower (visible) position than the other at their respective guild.
        """
        if type(self) is not type(other):
            return NotImplemented
        
        return other._is_greater_than_same_type(self)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two guild widget channels are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `type(self)`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # position
        if self.position != other.position:
            return False
        
        return True
    
    
    def _is_greater_than_same_type(self, other):
        """
        Returns whether self is greater than the other guild widget.
            
        Helper method for ``.__gt__`` and ``.__lt__`.
        
        Parameters
        ----------
        other : `type(self)`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_greater_than : `bool`
        """
        self_position = self.position
        other_position = other.position
        
        if self_position > other_position:
            return True
        
        if self_position < other_position:
            return False
        
        return self.id > other.id

    def copy(self):
        """
        Copies the guild widget channel.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = self.id
        new.name = self.name
        new.position = self.position
        return new
    
    
    def copy_with(
        self,
        *,
        name = ...,
        position = ...,
        channel_id = ...,
    ):
        """
        Copies the guild widget channel with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The unique identifier number of the guild widget channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = self.id
        else:
            channel_id = validate_id(channel_id)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # position
        if position is ...:
            position = self.position
        else:
            position = validate_position(position)
        
        # Construct
        new = object.__new__(type(self))
        new.id = channel_id
        new.name = name
        new.position = position
        return new
    
    
    @property
    def mention(self):
        """
        The channel's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'<#{self.id}>'



__all__ = ()

from ..application import Entitlement, Subscription
from ..application_command import ApplicationCommand
from ..auto_moderation import AutoModerationRule
from ..bases import maybe_snowflake, maybe_snowflake_pair, maybe_snowflake_token_pair
from ..channel import Channel, ForumTag, PermissionOverwrite
from ..core import (
    APPLICATION_COMMANDS, AUTO_MODERATION_RULES, CHANNELS, EMBEDDED_ACTIVITIES, EMOJIS, ENTITLEMENTS, FORUM_TAGS,
    GUILDS, MESSAGES, ROLES, SCHEDULED_EVENTS, SOUNDBOARD_SOUNDS, STICKERS, STICKER_PACKS, SUBSCRIPTIONS, USERS
)
from ..embedded_activity import EmbeddedActivity
from ..emoji import Emoji, Reaction, ReactionType
from ..guild import Guild
from ..message import Message
from ..oauth2 import Achievement, Oauth2Access, Oauth2User
from ..role import Role
from ..poll import PollAnswer
from ..scheduled_event import ScheduledEvent
from ..soundboard import SoundboardSound
from ..stage import Stage
from ..sticker import Sticker, StickerPack
from ..user import ClientUserBase
from ..webhook import Webhook


def validate_message_to_delete(message):
    """
    Validates a message to delete.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message to validate for deletion.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier where the message is.
    message_id : `int`
        The message's identifier.
    message : `None`, ``Message``
        The referenced message if found.
    
    Raises
    ------
    TypeError
        If message was not given neither as ``Message``, `tuple` (`int`, `int`).
    """
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {type(message).__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
        
        message = MESSAGES.get(message, None)
    
    return channel_id, message_id, message


def get_channel_id(channel, type_checker):
    """
    Gets the channel's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `int`
        The channel, or it's identifier.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, Channel):
        if type_checker(channel) or channel.partial:
            return channel.id
    
    else:
        channel_id = maybe_snowflake(channel)
        if (channel_id is not None):
            return channel_id
        
    raise TypeError(
        f'`channel` can be `{Channel.__name__}`, `int`,  passing the `{type_checker.__name__}` check, '
        f'got {channel.__class__.__name__}; {channel!r}.'
    )


def get_channel_guild_id_and_id(channel, type_checker):
    """
    Gets the channel's and it's guild's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `tuple` (`int`, `int`)
        The role, or `guild-id`, `role-id` pair.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    snowflake_pair : `tuple` (`int`, `int`)
        The channel's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, Channel):
        if type_checker(channel) or channel.partial:
            return channel.guild_id, channel.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(channel)
        if (snowflake_pair is not None):
            return snowflake_pair
    
    raise TypeError(
        f'`channel` can be `{Channel.__name__}`, `tuple` (`int`, `int`),  passing the `{type_checker.__name__}` check, '
        f'got {channel.__class__.__name__}; {channel!r}.'
    )


def build_get_channel_and_id_error_message(channel, type_checker):
    """
    Builds error message for `get_channel_and_id`.
    
    Parameters
    ----------
    channel : `object`
        The object that was passed as `channel`.
    type_checker : `None | FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    error_message : `str`
    """
    parts = ['`channel` can be `', Channel.__name__, '`, `int`']
    
    if (type_checker is not None):
        parts.append(', passing the `')
        parts.append(type_checker.__name__)
        parts.append('` check')
    
    parts.append(', got ')
    parts.append(type(channel).__name__)
    parts.append('; ')
    parts.append(repr(channel))
    parts.append('.')
    
    return ''.join(parts)


def get_channel_and_id(channel, type_checker = None):
    """
    Gets thread channel and it's identifier from the given thread channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `int`
        The channel, or it's identifier.
    type_checker : `None | FunctionType` = `None`, Optional
        Type checker for `channel`.
    
    Returns
    -------
    channel : `None`, ``Channel``
        The channel if found.
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    while True:
        if isinstance(channel, Channel):
            if (type_checker is None) or type_checker(channel) or channel.partial:
                channel_id = channel.id
                break
        
        else:
            channel_id = maybe_snowflake(channel)
            if (channel_id is not None):
                try:
                    channel = CHANNELS[channel_id]
                except KeyError:
                    channel = None
                    break
                
                if (type_checker is None) or type_checker(channel) or channel.partial:
                    channel_id = channel.id
                    break
        
        raise TypeError(build_get_channel_and_id_error_message(channel, type_checker))
    
    return channel, channel_id


def get_stage_and_channel_id(stage):
    """
    Gets the stage an its channel's identifier.
    
    Parameters
    ----------
    stage : `Stage`, ``Channel``, `int`
        The stage, or it's identifier.
    
    Returns
    -------
    stage : `None`, ``Stage``
        The identified stage if any.
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `stage`'s type is incorrect.
    """
    while True:
        if isinstance(stage, Stage):
            channel_id = stage.channel.id
            break
        
        elif isinstance(stage, Channel):
            if stage.is_guild_stage() or stage.partial:
                channel_id = stage.id
                stage = None
                break
        
        else:
            channel_id = maybe_snowflake(stage)
            if (channel_id is not None):
                stage = None
                break
        
        raise TypeError(
            f'`stage` can be `{Stage.__name__}`, `{Channel.__name__}`, `int`, got '
            f'{stage.__class__.__name__}; {stage!r}.'
        )
    
    return stage, channel_id


def get_stage_channel_id(stage):
    """
    Gets stage's channel's identifier from the given stage or of it's identifier.
    
    Parameters
    ----------
    stage : `Stage`, ``Channel``, `int`
        The stage, or it's identifier.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `stage`'s type is incorrect.
    """
    while True:
        if isinstance(stage, Stage):
            channel_id = stage.channel.id
            break
        
        elif isinstance(stage, Channel):
            if stage.is_guild_stage() or stage.partial:
                channel_id = stage.id
                break
        
        else:
            channel_id = maybe_snowflake(stage)
            if (channel_id is not None):
                break
        
        raise TypeError(
            f'`stage` can be `{Stage.__name__}`, `{Channel.__name__}`, `int`, got '
            f'{stage.__class__.__name__}; {stage!r}.'
        )
    
    return channel_id


def get_user_id(user):
    """
    Gets user identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if isinstance(user, ClientUserBase):
        return user.id
    
    user_id = maybe_snowflake(user)
    if (user_id is not None):
        return user_id
    
    raise TypeError(
        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
    )


def get_user_and_id(user):
    """
    Gets user and it's identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user : `None`, ``ClientUserBase``
        The user if found.
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    while True:
        if isinstance(user, ClientUserBase):
            user_id = user.id
            break
        
        user_id = maybe_snowflake(user)
        if (user_id is not None):
            try:
                user = USERS[user_id]
            except KeyError:
                user = None
                break
            
            if isinstance(user, ClientUserBase):
                break
        
        raise TypeError(
            f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
        )
        
    return user, user_id


def get_user_id_nullable(user):
    """
    Gets user identifier from the given user or of it's identifier.
    
    > The user can be `None`. At that case `user_id` will default to `0`.
    
    Parameters
    ----------
    user : `None`, ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if user is None:
        return 0
    
    if isinstance(user, ClientUserBase):
        return user.id
    
    user_id = maybe_snowflake(user)
    if (user_id is not None):
        return user_id
        
    raise TypeError(
        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
    )


def get_guild_id(guild):
    """
    Gets the guild's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        return guild.id
    
    guild_id = maybe_snowflake(guild)
    if (guild_id is not None):
        return guild_id
        
    raise TypeError(
        f'`guild` can be `{Guild.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
    )


def get_guild_and_id(guild):
    """
    Gets the guild and it's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild : `None`, ``Guild``
        The guild if found.
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
    else:
        guild_id = maybe_snowflake(guild)
        if guild_id is None:
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        guild = GUILDS.get(guild_id, None)
    
    return guild, guild_id


def get_achievement_id(achievement):
    """
    Gets the achievement identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(
                f'`achievement` can be `{Achievement.__name__}`, `int`, got '
                f'{type(achievement).__name__}; {achievement!r}.'
            )
    
    return achievement_id


def get_achievement_and_id(achievement):
    """
    Gets the achievement and it's identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement : ``Achievement``, `None`
        The achievement if found.
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(
                f'`achievement` can be `{Achievement.__name__}`, `int`, got '
                f'{achievement.__class__.__name__}; {achievement!r}.'
            )
        
        achievement = None
    
    return achievement, achievement_id


def get_channel_id_and_message_id(message):
    """
    Gets the message's channel's and it's own identifier.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message or it' representation.
    
    Returns
    -------
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    # 1.: Message
    # 4.: None -> raise
    # 5.: `tuple` (`int`, `int`)
    # 6.: raise
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    elif message is None:
        raise TypeError(
            '`message` was given as `None`. Make sure to call message create methods with non-empty content(s).'
        )
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
    
    return channel_id, message_id


def get_message_and_channel_id_and_message_id(message):
    """
    Gets the message's channel's and it's own identifier.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message or it' representation.
    
    Returns
    -------
    message : `None`, ``Message``
        The message in context if found.
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    # 1.: Message
    # 4.: None -> raise
    # 5.: `tuple` (`int`, `int`)
    # 6.: raise
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    elif message is None:
        raise TypeError(
            '`message` was given as `None`. Make sure to call message create methods with non-empty content(s).'
        )
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
        message = MESSAGES.get(message_id, None)
    
    return message, channel_id, message_id


def get_role_id(role):
    """
    Gets the role identifier from the given role or of it's identifier.
    
    Parameters
    ----------
    role : ``Role``, `int`
        The role, or it's identifier.
    
    Returns
    -------
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        role_id = role.id
    else:
        role_id = maybe_snowflake(role)
        if role_id is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `int`, got {role.__class__.__name__}; {role!r}.'
            )
    
    return role_id


def get_role_role_guild_id_and_id(role):
    """
    Gets the role identifier from the given role or of it's identifier.
    
    Parameters
    ----------
    role : ``Role``, `tuple` (`int`, `int`)
        The role, or a `guild-id`, `role-id` pair.
    
    Returns
    -------
    role : `None`, ``Role``
        The respective role.
    
    guild_id : `int`
        The role's guild's identifier.
    
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        role_id = role.id
        guild_id = role.guild_id
    
    else:
        snowflake_pair = maybe_snowflake_pair(role)
        if snowflake_pair is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got {role.__class__.__name__}; {role!r}.'
            )
        
        guild_id, role_id = snowflake_pair
        role = ROLES.get(role_id, None)
    
    return role, guild_id, role_id


def get_role_guild_id_and_id(role):
    """
    Gets the role's and it's guild's identifier from the given role or of a `guild-id`, `role-id` pair.
    
    Parameters
    ----------
    role : ``Role``, `tuple` (`int`, `int`)
        The role, or a `guild-id`, `role-id` pair.
    
    Returns
    -------
    guild_id : `int`
        The role's guild's identifier.
    
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        return role.guild_id, role.id
    
    snowflake_pair = maybe_snowflake_pair(role)
    if snowflake_pair is not None:
        return snowflake_pair
    
    raise TypeError(
        f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got {role.__class__.__name__}; {role!r}.'
    )


def get_webhook_id(webhook):
    """
    Gets the webhook's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        webhook_id = webhook.id
    else:
        webhook_id = maybe_snowflake(webhook)
        if webhook_id is None:
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `int`, got {webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return webhook_id


def get_webhook_and_id(webhook):
    """
    Gets the webhook and it's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook : ``Webhook``, `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            break
        
        webhook_id = maybe_snowflake(webhook)
        if (webhook_id is not None):
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
            
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `int`, got {webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return webhook, webhook_id


def get_webhook_id_and_token(webhook):
    """
    Gets the webhook's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        snowflake_token_pair = webhook.id, webhook.token
    else:
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is None):
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `tuple` (`int`, `str`), got '
                f'{webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return snowflake_token_pair


def get_webhook_and_id_and_token(webhook):
    """
    Gets the webhook, it's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook : ``Webhook``, `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            webhook_token = webhook.token
            break
        
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is not None):
            webhook_id, webhook_token = snowflake_token_pair
            
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
        
        raise TypeError(
            f'`webhook` can be `{Webhook.__name__}`, `tuple` (`int`, `str`), got '
            f'{webhook.__class__.__name__}; {webhook!r}.'
        )
    
    return webhook, webhook_id, webhook_token


def get_reaction_emoji_value_and_type(reaction):
    """
    Gets the reaction representation form of the given reaction.
    
    Parameters
    ----------
    reaction : ``Reaction``, ``Emoji``, `str`
        The emoji to get it's reaction form of.
    
    Returns
    -------
    emoji_value : `str`
        The emoji's reaction form.
    reaction_type : ``ReactionType``
        The reaction's type.
    
    Raises
    ------
    TypeError
        If `reaction`'s type is incorrect.
    """
    if isinstance(reaction, Emoji):
        return reaction.as_reaction, ReactionType.standard
    
    if isinstance(reaction, Reaction):
        return reaction.emoji.as_reaction, reaction.type
    
    if isinstance(reaction, str):
        return reaction, ReactionType.standard
    
    raise TypeError(
        f'`reaction` can be `{Reaction.__name__}`, `{Emoji.__name__}`, `str`, '
        f'got {type(reaction).__name__}; {reaction!r}.'
    )


def get_emoji_guild_id_and_id(emoji):
    """
    Gets the emoji's guild's and its identifier from the given value.
    
    Parameters
    ----------
    emoji : ``Emoji``, `(int, int)`
        The emoji or `guild-id`, `emoji-id` pair.
    
    Returns
    -------
    snowflake_pair : `(int, int)`
        The emoji's guild's and it's own identifier.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    """
    if isinstance(emoji, Emoji):
        snowflake_pair = emoji.guild_id, emoji.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(emoji)
        if snowflake_pair is None:
            raise TypeError(
                f'`emoji` can be `{Emoji.__name__}`, `(int, int)`, got {type(emoji).__name__}; {emoji!r}.'
            )
    
    return snowflake_pair


def get_emoji_id(emoji):
    """
    Gets the emoji's identifier.
    
    Parameters
    ----------
    emoji : ``Emoji``, `int`
        The emoji or its identifier.
    
    Returns
    -------
    emoji_id : `int`
    """
    if isinstance(emoji, Emoji):
        emoji_id = emoji.id
    else:
        emoji_id = maybe_snowflake(emoji)
        if (emoji_id is None):
            raise TypeError(
                f'`emoji` can be `{Emoji.__name__}`, `int`, got {type(emoji).__name__}; {emoji!r}.'
            )
    
    return emoji_id


def get_emoji_and_id(emoji):
    """
    Gets the emoji and its identifier.
    
    Parameters
    ----------
    emoji : ``Emoji``, `int`
        The emoji or its identifier.
    
    Returns
    -------
    emoji_and_emoji_id : `(None | Emoji, int)`
    """
    if isinstance(emoji, Emoji):
        emoji_id = emoji.id
    else:
        emoji_id = maybe_snowflake(emoji)
        if (emoji_id is None):
            raise TypeError(
                f'`emoji` can be `{Emoji.__name__}`, `int`, got {type(emoji).__name__}; {emoji!r}.'
            )
        emoji = EMOJIS.get(emoji_id, None)
    
    return emoji, emoji_id


def get_emoji_and_guild_id_and_id(emoji):
    """
    Gets the emoji and it's guild's identifier from the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji``, `(int, int)`
        The emoji or `guild-id`, `emoji-id` pair.
    
    Returns
    -------
    emoji_and_guild_id_and_emoji_id : `(None | Emoji, int, int)`
        The emoji and its guild's and it's own identifier.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    """
    if isinstance(emoji, Emoji):
        guild_id = emoji.guild_id
        emoji_id = emoji.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(emoji)
        if snowflake_pair is None:
            raise TypeError(
                f'`emoji` can be `{Emoji.__name__}`, `(int, int)`, got {type(emoji).__name__}; {emoji!r}.'
            )
        
        guild_id, emoji_id = snowflake_pair
        emoji = EMOJIS.get(emoji_id, None)
    
    return emoji, guild_id, emoji_id


def get_sticker_and_id(sticker):
    """
    Gets sticker and it's identifier from the given sticker or of it's identifier.
    
    Parameters
    ----------
    sticker : ``Sticker``, `int`
        The sticker, or it's identifier.
    
    Returns
    -------
    sticker : ``Sticker``, `None`
        The sticker if found.
    sticker_id : `int`
        The sticker's identifier.
    
    Raises
    ------
    TypeError
        If `sticker`'s type is incorrect.
    """
    while True:
        if isinstance(sticker, Sticker):
            sticker_id = sticker.id
            break
        
        sticker_id = maybe_snowflake(sticker)
        if (sticker_id is not None):
            sticker = STICKERS.get(sticker_id, None)
            break
        
        raise TypeError(
            f'`sticker` can be `{Sticker.__name__}`, `int`, got {sticker.__class__.__name__}; {sticker!r}.'
        )
        
    return sticker, sticker_id


def get_sticker_pack_and_id(sticker_pack):
    """
    Gets sticker pack and it's identifier from the given sticker pack or of it's identifier.
    
    Parameters
    ----------
    sticker_pack : ``StickerPack``, `int`
        The sticker, or it's identifier.
    
    Returns
    -------
    sticker_pack : ``StickerPack``, `None`
        The sticker pack if found.
    sticker_pack_id : `int`
        The sticker pack's identifier.
    
    Raises
    ------
    TypeError
        If `sticker_pack`'s type is incorrect.
    """
    while True:
        if isinstance(sticker_pack, StickerPack):
            sticker_pack_id = sticker_pack.id
            break
        
        sticker_pack_id = maybe_snowflake(sticker_pack)
        if (sticker_pack_id is not None):
            sticker_pack = STICKER_PACKS.get(sticker_pack_id, None)
            break
        
        raise TypeError(
            f'`sticker_pack` can be `{StickerPack.__name__}`, `int`, got {sticker_pack.__class__.__name__}; '
            f'{sticker_pack!r}.'
        )
        
    return sticker_pack, sticker_pack_id


def get_scheduled_event_guild_id_and_id(scheduled_event):
    """
    Gets the scheduled event's and it's identifier from the given scheduled event or from a tuple of 2 identifiers.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `tuple` of (`int`, `int`)
        The scheduled event, or a tuple of two identifiers.
    
    Returns
    -------
    guild_id : `int`
        The scheduled event's guild's identifier.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        return scheduled_event.guild_id, scheduled_event.id
    
    
    snowflake_pair = maybe_snowflake_pair(scheduled_event)
    if snowflake_pair is None:
        raise TypeError(
            f'`scheduled_event` can be `{ScheduledEvent.__name__}`, `tuple` of (`int`, `int`), got '
            f'{scheduled_event.__class__.__name__}; {scheduled_event!r}.'
        )
    
    return snowflake_pair


def get_scheduled_event_and_guild_id_and_id(scheduled_event):
    """
    Gets the scheduled event, it's guild's identifier and it's identifier.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `tuple` (`int`, `int`)
        The scheduled event, or it's identifier.
    
    Returns
    -------
    scheduled_event : `None`, ``ScheduledEvent``
        The scheduled event.
    guild_id : `int`
        The scheduled event's guild's identifier.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        scheduled_event =  scheduled_event
        scheduled_event_id = scheduled_event.id
        guild_id = scheduled_event.guild_id
    else:
        snowflake_pair = maybe_snowflake_pair(scheduled_event)
        if snowflake_pair is None:
            raise TypeError(
                f'`scheduled_event` can be `{ScheduledEvent.__name__}`, `tuple` of (`int`, `int`), got '
                f'{scheduled_event.__class__.__name__}; {scheduled_event!r}.'
            )
        
        guild_id, scheduled_event_id = snowflake_pair
        
        scheduled_event = SCHEDULED_EVENTS.get(scheduled_event_id, None)
    
    return scheduled_event, guild_id, scheduled_event_id


def get_application_command_id(application_command):
    """
    Gets the application command's identifier.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
    
    return application_command_id


def get_application_command_and_id(application_command):
    """
    Gets the application command's identifier.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command : `None`, ``ApplicationCommand``
        The application command if exists.
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
        
        application_command = APPLICATION_COMMANDS.get(application_command_id, None)
    
    return application_command, application_command_id


def get_application_command_id_nullable(application_command):
    """
    Gets the application command's identifier.
    
    > The application command can be `None`. At that case `application_command_id` will default to `0`.
    
    Parameters
    ----------
    application_command : `None`, ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if application_command is None:
        application_command_id = 0
    
    elif isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
    
    return application_command_id


def get_auto_moderation_rule_guild_id_and_id(auto_moderation_rule):
    """
    Gets the auto moderation rule's guild's identifier and it's own identifier.
    
    Parameters
    ----------
    auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
        The auto moderation rule, or it's identifier as a `guild-id`, `rule-id` pair..
    
    Returns
    -------
    guild_id : `int`
        The auto moderation rule's guild's identifier.
    auto_moderation_rule_id : `int`
        The auto moderation rule's identifier.
    
    Raises
    ------
    TypeError
        If `auto_moderation_rule`'s type is incorrect.
    """
    if isinstance(auto_moderation_rule, AutoModerationRule):
        guild_id = auto_moderation_rule.guild_id
        rule_id = auto_moderation_rule.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(auto_moderation_rule)
        if snowflake_pair is None:
            raise TypeError(
                f'`auto_moderation_rule` can be `{AutoModerationRule.__name__}`, `tuple` of (`int`, `int`), got '
                f'{auto_moderation_rule.__class__.__name__}; {auto_moderation_rule!r}.'
            )
        
        guild_id, rule_id = snowflake_pair
    
    return guild_id, rule_id


def get_auto_moderation_rule_and_guild_id_and_id(auto_moderation_rule):
    """
    Gets the auto moderation, its guild's identifier and it's own identifier.
    
    Parameters
    ----------
    auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
        The auto moderation rule, or it's identifier as a `guild-id`, `rule-id` pair..
    
    Returns
    -------
    auto_moderation_rule : `None`, ``AutoModerationRule``
        The auto moderation rule.
    guild_id : `int`
        The auto moderation rule's guild's identifier.
    auto_moderation_rule_id : `int`
        The auto moderation rule's identifier.
    
    Raises
    ------
    TypeError
        If `auto_moderation_rule`'s type is incorrect.
    """
    if isinstance(auto_moderation_rule, AutoModerationRule):
        guild_id = auto_moderation_rule.guild_id
        rule_id = auto_moderation_rule.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(auto_moderation_rule)
        if snowflake_pair is None:
            raise TypeError(
                f'`auto_moderation_rule` can be `{AutoModerationRule.__name__}`, `tuple` of (`int`, `int`), got '
                f'{auto_moderation_rule.__class__.__name__}; {auto_moderation_rule!r}.'
            )
        
        guild_id, rule_id = snowflake_pair
        auto_moderation_rule = AUTO_MODERATION_RULES.get(rule_id, None)
    
    return auto_moderation_rule, guild_id, rule_id


def get_forum_tag_id(forum_tag):
    """
    Gets the forum tag's identifier.
    
    Parameters
    ----------
    forum_tag : ``ForumTag``, `int`
        The forum tag or its identifier.
    
    Returns
    -------
    forum_tag_id : `int`
        The forum tag's identifier.
    
    Raises
    ------
    TypeError
        - If `forum_tag` type is incorrect.
    """
    if isinstance(forum_tag, ForumTag):
        forum_tag_id = forum_tag.id
    
    else:
        forum_tag_id = maybe_snowflake(forum_tag)
        if (forum_tag_id is None):
            raise TypeError(
                f'`forum_tag` can be `{ForumTag.__name__}, `int`, got {forum_tag.__class__.__name__}; {forum_tag!r}.'
            )
    
    return forum_tag_id


def get_forum_tag_and_id(forum_tag):
    """
    Gets the forum tag and it's identifier.
    
    Parameters
    ----------
    forum_tag : ``ForumTag``, `int`
        The forum tag or its identifier.
    
    Returns
    -------
    forum_tag : `None`, ``ForumTag``
        The forum tag if exists.
    forum_tag_id : `int`
        The forum tag's identifier.
    
    Raises
    ------
    TypeError
        - If `forum_tag` type is incorrect.
    """
    if isinstance(forum_tag, ForumTag):
        forum_tag_id = forum_tag.id
    
    else:
        forum_tag_id = maybe_snowflake(forum_tag)
        if (forum_tag_id is None):
            raise TypeError(
                f'`forum_tag` can be `{ForumTag.__name__}, `int`, got {forum_tag.__class__.__name__}; {forum_tag!r}.'
            )
        
        forum_tag = FORUM_TAGS.get(forum_tag_id, None)
    
    return forum_tag, forum_tag_id


def get_permission_overwrite_target_id(permission_overwrite):
    """
    Returns the given `permission_overwrite`'s target id.
    
    Parameters
    ----------
    permission_overwrite : ``PermissionOverwrite``
        The permission overwrite or it's target id.
    
    Returns
    -------
    target_id : `int`
    
    Raises
    ------
    TypeError
        - If `permission_overwrite` type is incorrect.
    """
    if isinstance(permission_overwrite, PermissionOverwrite):
        target_id = permission_overwrite.target_id
    
    else:
        target_id = maybe_snowflake(permission_overwrite)
        if (target_id is None):
            raise TypeError(
                f'`permission_overwrite` can be `{PermissionOverwrite.__name__}`, `int` got '
                f'{permission_overwrite.__class__.__name__}; {permission_overwrite!r}.'
            )
    
    return target_id


def _check_required_scope(access, required_scope):
    if (required_scope is not None) and (not access.has_scope(required_scope)):
        raise ValueError(
            f'The given `access` not grants {required_scope.name!r} scope, what is required, '
            f'got {access!r}.'
        )


def get_oauth2_access_token(access, required_scope = None):
    """
    Returns the given oauth2 access's access token. Accepts both access and just token as well.
    
    Parameters
    ----------
    access : ``Oauth2Access``, ``Oauth2User``, `str`
        Oauth2 access to the respective user or it's access token.
    required_scope : `None`, ``Oauth2Scope`` = `None`, Optional
        Required scope of the access if any.
    
    Returns
    -------
    access_token : `str`
    
    Raises
    ------
    TypeError
        - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
    ValueError
        - If the given `access` is not providing the required scope.
    """
    if isinstance(access, (Oauth2Access, Oauth2User)):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
    
    elif isinstance(access, str):
        access_token = access
    
    else:
        raise TypeError(
            f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`, `str`'
            f', got {access.__class__.__name__}; {access!r}.'
        )
    
    return access_token


def get_oauth2_access_token_and_user_id(access, user, required_scope = None):
    """
    Returns the given oauth2 access's access token and the user's identifier.
    
    Parameters
    ----------
    access : ``Oauth2Access``, ``Oauth2User``, `str`
        Oauth2 access to the respective user or it's access token.
    required_scope : `None`, ``Oauth2Scope`` = `None`, Optional
        Required scope of the access if any.
    
    Raises
    ------
    TypeError
        - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        - If `user` is not `None`, ``ClientUserBase``.
        - If `user.id` could not be determined.
    ValueError
        - If the given `access` is not providing the required scope.
        - If `user` and `access` refers to a different user.
    """
    user_id = get_user_id_nullable(user)
    
    if isinstance(access, Oauth2Access):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
    
    elif isinstance(access, Oauth2User):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
        
        if user_id and (user_id != access.id):
            raise ValueError(
                f'The given `user` and `access` refers to different users, got user = {user!r}, '
                f'access = {access!r}.'
            )
        
        user_id = access.id
    
    elif isinstance(access, str):
        access_token = access
    
    else:
        raise TypeError(
            f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`, `str`, got '
            f'{access.__class__.__name__}; {access!r}.'
        )
    
    
    if not user_id:
        raise TypeError(
            f'`user` was not detectable neither from `user` nor from `access` parameters, got '
            f'user = {user!r}, access = {access!r}.'
        )
    
    return access_token, user_id


def get_soundboard_sound_and_guild_id_and_id(soundboard_sound):
    """
    Gets the soundboard sound identifier from the given soundboard sound or of it's identifier.
    
    Parameters
    ----------
    soundboard_sound : ``SoundboardSound``, `tuple` (`int`, `int`)
        The soundboard sound, or a `guild-id`, `soundboard-sound-id` pair.
    
    Returns
    -------
    soundboard_sound : `None`, ``SoundboardSound``
        The respective soundboard sound.
    
    guild_id : `int`
        The soundboard sound's guild's identifier.
    
    soundboard_sound_id : `int`
        The soundboard sound's identifier.
    
    Raises
    ------
    TypeError
        If `soundboard_sound`'s type is incorrect.
    """
    if isinstance(soundboard_sound, SoundboardSound):
        soundboard_sound_id = soundboard_sound.id
        guild_id = soundboard_sound.guild_id
    
    else:
        snowflake_pair = maybe_snowflake_pair(soundboard_sound)
        if snowflake_pair is None:
            raise TypeError(
                f'`soundboard_sound` can be `{SoundboardSound.__name__}`, `tuple` (`int`, `int`), '
                f'got {soundboard_sound.__class__.__name__}; {soundboard_sound!r}.'
            )
        
        guild_id, soundboard_sound_id = snowflake_pair
        soundboard_sound = SOUNDBOARD_SOUNDS.get(soundboard_sound_id, None)
    
    return soundboard_sound, guild_id, soundboard_sound_id


def get_soundboard_sound_guild_id_and_id(soundboard_sound):
    """
    Gets the soundboard_sound's and it's guild's identifier from the given soundboard_sound or of a `guild-id`,
    `soundboard_sound-id` pair.
    
    Parameters
    ----------
    soundboard_sound : ``SoundboardSound``, `tuple` (`int`, `int`)
        The soundboard sound, or a `guild-id`, `soundboard-sound-id` pair.
    
    Returns
    -------
    guild_id : `int`
        The soundboard sound's guild's identifier.
    
    soundboard_sound_id : `int`
        The soundboard sound's identifier.
    
    Raises
    ------
    TypeError
        If `soundboard_sound`'s type is incorrect.
    """
    if isinstance(soundboard_sound, SoundboardSound):
        return soundboard_sound.guild_id, soundboard_sound.id
    
    snowflake_pair = maybe_snowflake_pair(soundboard_sound)
    if snowflake_pair is not None:
        return snowflake_pair
    
    raise TypeError(
        f'`soundboard_sound` can be `{SoundboardSound.__name__}`, `tuple` (`int`, `int`), '
        f'got {soundboard_sound.__class__.__name__}; {soundboard_sound!r}.'
    )


def get_entitlement_id(entitlement):
    """
    Gets the entitlement's identifier.
    
    Parameters
    ----------
    entitlement : ``Entitlement``, `int`
        The entitlement or its identifier.
    
    Returns
    -------
    entitlement_id : `int`
        The entitlement's identifier.
    
    Raises
    ------
    TypeError
        - If `entitlement` type is incorrect.
    """
    if isinstance(entitlement, Entitlement):
        entitlement_id = entitlement.id
    
    else:
        entitlement_id = maybe_snowflake(entitlement)
        if (entitlement_id is None):
            raise TypeError(
                f'`entitlement` can be `{Entitlement.__name__}, `int`, '
                f'got {entitlement.__class__.__name__}; {entitlement!r}.'
            )
    
    return entitlement_id


def get_entitlement_and_id(entitlement):
    """
    Gets the entitlement and its identifier.
    
    Parameters
    ----------
    entitlement : `Entitlement | int`
        The entitlement to get its identifier.
    
    Returns
    -------
    entitlement : `None | Entitlement`
    entitlement_id : `int`
    
    Raises
    ------
    TypeError
        - If `entitlement` type is incorrect.
    """
    if isinstance(entitlement, Entitlement):
        return entitlement, entitlement.id
    
    entitlement_id = maybe_snowflake(entitlement)
    if entitlement_id is None:
        raise TypeError(
            f'`entitlement` can be `{Entitlement.__name__}, `int`, '
            f'got {type(entitlement).__name__}; {entitlement!r}.'
        )
    
    entitlement = ENTITLEMENTS.get(entitlement_id, None)
    return entitlement, entitlement_id


def get_poll_answer_and_id(poll_answer):
    """
    Gets the poll answer and its identifier.
    
    Parameters
    ----------
    poll_answer : ``PollAnswer``, `int`
        The poll answer to get its identifier.
    
    Returns
    -------
    poll_answer : `None`, ``PollAnswer``
    poll_answer_id : `int`
    
    Raises
    ------
    TypeError
        - If `poll_answer` type is incorrect.
    """
    if isinstance(poll_answer, PollAnswer):
        return poll_answer, poll_answer.id
    
    poll_answer_id = maybe_snowflake(poll_answer)
    if poll_answer_id is None:
        raise TypeError(
            f'`poll_answer` can be `{PollAnswer.__name__}, `int`, '
            f'got {poll_answer.__class__.__name__}; {poll_answer!r}.'
        )
    
    return None, poll_answer_id


def get_message_id(message):
    """
    Gets the message's identifier from the given message or of it's identifier.
    
    Parameters
    ----------
    message : ``Message``, `int`
        The message, or it's identifier.
    
    Returns
    -------
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    if isinstance(message, Message):
        message_id = message.id
    else:
        message_id = maybe_snowflake(message)
        if message_id is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, `int`, got {type(message).__name__}; {message!r}.'
            )
    
    return message_id


def get_embedded_activity_and_id(embedded_activity):
    """
    Gets the embedded activity and its identifier.
    
    Parameters
    ----------
    embedded_activity : `EmbeddedActivity | int`
        The embedded activity to get its identifier.
    
    Returns
    -------
    embedded_activity : `None | EmbeddedActivity`
    embedded_activity_id : `int`
    
    Raises
    ------
    TypeError
        - If `embedded_activity` type is incorrect.
    """
    if isinstance(embedded_activity, EmbeddedActivity):
        return embedded_activity, embedded_activity.id
    
    embedded_activity_id = maybe_snowflake(embedded_activity)
    if embedded_activity_id is None:
        raise TypeError(
            f'`embedded_activity` can be `{EmbeddedActivity.__name__}, `int`, '
            f'got {type(embedded_activity).__name__}; {embedded_activity!r}.'
        )
    
    embedded_activity = EMBEDDED_ACTIVITIES.get(embedded_activity_id, None)
    return embedded_activity, embedded_activity_id


def get_subscription_and_sku_id_and_id(subscription):
    """
    Gets the subscription and it's sku's identifier from the given subscription.
    
    Parameters
    ----------
    subscription : ``Subscription``, `(int, int)`
        The subscription or `sku-id`, `subscription-id` pair.
    
    Returns
    -------
    subscription_and_sku_id_and_subscription_id : `(None | Subscription, int, int)`
        The subscription and its sku's and it's own identifier.
    
    Raises
    ------
    TypeError
        If `subscription`'s type is incorrect.
    """
    if isinstance(subscription, Subscription):
        sku_id = subscription.sku_id
        subscription_id = subscription.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(subscription)
        if snowflake_pair is None:
            raise TypeError(
                f'`subscription` can be `{Subscription.__name__}`, `(int, int)`, got {type(subscription).__name__}; {subscription!r}.'
            )
        
        sku_id, subscription_id = snowflake_pair
        subscription = SUBSCRIPTIONS.get(subscription_id, None)
    
    return subscription, sku_id, subscription_id



__all__ = ('OnboardingPrompt',)

from ...bases import DiscordEntity
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_id, parse_in_onboarding, parse_name, parse_options, parse_required, parse_single_select, parse_type,
    put_id, put_in_onboarding, put_name, put_options, put_required, put_single_select,
    put_type, validate_id, validate_in_onboarding, validate_name, validate_options, validate_required,
    validate_single_select, validate_type
)
from .preinstanced import OnboardingPromptType


PRECREATE_FIELDS = {
    'in_onboarding': ('in_onboarding', validate_in_onboarding),
    'name': ('name', validate_name),
    'options': ('options', validate_options),
    'required': ('required', validate_required),
    'single_select': ('single_select', validate_single_select),
}


class OnboardingPrompt(DiscordEntity):
    """
    Option of an onboarding prompt.
    
    Attributes
    ----------
    in_onboarding : `bool`
        Whether this prompt is in the onboarding flow.
    id : `int`
        The prompt's identifier.
    name : `str`
        The prompt's name
    options : `None`, `tuple` of ``OnboardingPromptOption``
        The options of the prompt.
    required : `bool`
        Whether this prompt is required to do in the onboarding flow.
    single_select : `bool`
        Whether only one option can be selected.
    type : ``OnboardingPromptType``
        The prompt's type.
    """
    __slots__ = ('in_onboarding', 'name', 'options', 'required', 'single_select', 'type')
    
    def __new__(
        cls, *, in_onboarding = ..., name = ..., options = ..., prompt_type = ..., required = ..., single_select = ...
    ):
        """
        Creates an onboarding prompt instance from the given parameters.
        
        Parameters
        ----------
        in_onboarding : `bool`, Optional (Keyword only)
            Whether this prompt is in the onboarding flow.
        name : `str`, Optional (Keyword only)
            The prompt's name
        options : `None`, `iterable` of ``OnboardingPromptOption``, Optional (Keyword only)
            The options of the prompt.
        prompt_type : ``OnboardingPromptType``, `int`, Optional (Keyword only)
            The prompt's type.
        required : `bool`, Optional (Keyword only)
            Whether this prompt is required to do in the onboarding flow.
        single_select : `bool`, Optional (Keyword only)
            Whether only one option can be selected.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # in_onboarding
        if in_onboarding is ...:
            in_onboarding = False
        else:
            in_onboarding = validate_in_onboarding(in_onboarding)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # options
        if options is ...:
            options = None
        else:
            options = validate_options(options)
        
        # prompt_type
        if prompt_type is ...:
            prompt_type = OnboardingPromptType.multiple_choice
        else:
            prompt_type = validate_type(prompt_type)
        
        # required
        if required is ...:
            required = False
        else:
            required = validate_required(required)
        
        # single_select
        if single_select is ...:
            single_select = False
        else:
            single_select = validate_single_select(single_select)
        
        self = object.__new__(cls)
        self.in_onboarding = in_onboarding
        self.id = 0
        self.name = name
        self.options = options
        self.required = required
        self.single_select = single_select
        self.type = prompt_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new onboarding prompt from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Onboarding prompt option data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.in_onboarding = parse_in_onboarding(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.options = parse_options(data)
        self.required = parse_required(data)
        self.single_select = parse_single_select(data)
        self.type = parse_type(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the onboarding prompt to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_in_onboarding(self.in_onboarding, data, defaults)
        put_name(self.name, data, defaults)
        put_options(self.options, data, defaults, include_internals = include_internals)
        put_required(self.required, data, defaults)
        put_single_select(self.single_select, data, defaults)
        put_type(self.type, data, defaults)
        
        if include_internals:
            put_id(self.id, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the onboarding prompt's representation."""
        repr_parts = [
            '<',
            type(self).__name__,
        ]
        
        prompt_id = self.id
        if prompt_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(prompt_id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        prompt_type = self.type
        repr_parts.append(', type = ')
        repr_parts.append(prompt_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(prompt_type.value))
        
        in_onboarding = self.in_onboarding
        if in_onboarding:
            repr_parts.append(', in_onboarding = ')
            repr_parts.append(repr(in_onboarding))
        
        required = self.required
        if required:
            repr_parts.append(', required = ')
            repr_parts.append(repr(required))
            
        single_select = self.single_select
        if single_select:
            repr_parts.append(', single_select = ')
            repr_parts.append(repr(single_select))
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options = [')
            
            index = 0
            length = len(options)
            
            while True:
                option = options[index]
                index += 1
                
                repr_parts.append(repr(option))
                
                if index == length:
                    break
                
                repr_parts.append(', ')
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two onboarding prompts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two onboarding prompts are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two onboarding prompts are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        are_equal : `bool`
        """
        if self.in_onboarding != other.in_onboarding:
            return False
        
        # id
        # Ignore it
        
        if self.name != other.name:
            return False
        
        if self.options != other.options:
            return False
        
        if self.required != other.required:
            return False
        
        if self.single_select != other.single_select:
            return False
        
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the onboarding prompt's hash value."""
        hash_value = 0
        
        hash_value ^= self.in_onboarding
        
        # id
        # ignore it
        
        name = self.name
        if (name is not None):
            hash_value ^= hash(name)
        
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 2
            
            for option in options:
                hash_value ^= hash(option)
        
        hash_value ^= self.required << 5
        
        hash_value ^= self.single_select << 6
        
        hash_value ^= self.type.value << 7
        
        return hash_value
    
    
    @classmethod
    def precreate(
        cls,
        prompt_id,
        prompt_type = ...,
        **keyword_parameters,
    ):
        """
        Precreates an onboarding prompt. Since they are not cached, this method just a ``.__new__`` alternative.
        
        Parameters
        ----------
        prompt_id : `int`
            The prompt's identifier.
        prompt_type : ``PromptType``, `int`, Optional (Keyword only)
            The prompt's type.
        **keyword_parameters : Keyword parameters
            Additional parameters defining how the option's fields should be set.
        
        Other Parameters
        ----------------
        in_onboarding : `bool`, Optional (Keyword only)
            Whether this prompt is in the onboarding flow.
        name : `str`, Optional (Keyword only)
            The prompt's name
        options : `None`, `iterable` of ``OnboardingPromptOption``, Optional (Keyword only)
            The options of the prompt.
        required : `bool`, Optional (Keyword only)
            Whether this prompt is required to do in the onboarding flow.
        single_select : `bool`, Optional (Keyword only)
            Whether only one option can be selected.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        prompt_id = validate_id(prompt_id)
        
        # prompt_type
        if prompt_type is ...:
            prompt_type = OnboardingPromptType.multiple_choice
        else:
            prompt_type = validate_type(prompt_type)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        # Construct
        
        self = object.__new__(cls)
        self.in_onboarding = False
        self.id = prompt_id
        self.name = ''
        self.options = None
        self.required = False
        self.single_select = False
        self.type = prompt_type
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def copy(self):
        """
        Copies the onboarding prompt.
        
        Returns
        -------
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.in_onboarding = self.in_onboarding
        new.id = 0
        new.name = self.name
        
        options = self.options
        if (options is not None):
            options = (*(option.copy() for option in options),)
        new.options = options
        
        new.required = self.required
        new.single_select = self.single_select
        new.type = self.type
        return new
    
    
    def copy_with(
        self, *, in_onboarding = ..., name = ..., options = ..., prompt_type = ..., required = ..., single_select = ...
    ):
        """
        Copies the onboarding prompt with the given fields.
        
        Parameters
        ----------
        in_onboarding : `bool`, Optional (Keyword only)
            Whether this prompt is in the onboarding flow.
        name : `str`, Optional (Keyword only)
            The prompt's name
        options : `None`, `iterable` of ``OnboardingPromptOption``, Optional (Keyword only)
            The options of the prompt.
        prompt_type : ``OnboardingPromptType``, `int`, Optional (Keyword only)
            The prompt's type.
        required : `bool`, Optional (Keyword only)
            Whether this prompt is required to do in the onboarding flow.
        single_select : `bool`, Optional (Keyword only)
            Whether only one option can be selected.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # in_onboarding
        if in_onboarding is ...:
            in_onboarding = self.in_onboarding
        else:
            in_onboarding = validate_in_onboarding(in_onboarding)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # options
        if options is ...:
            options = self.options
            if (options is not None):
                options = (*(option.copy() for option in options),)
        else:
            options = validate_options(options)
        
        # prompt_type
        if prompt_type is ...:
            prompt_type = self.type
        else:
            prompt_type = validate_type(prompt_type)
        
        # required
        if required is ...:
            required = self.required
        else:
            required = validate_required(required)
        
        # single_select
        if single_select is ...:
            single_select = self.single_select
        else:
            single_select = validate_single_select(single_select)
        
        new = object.__new__(type(self))
        new.in_onboarding = in_onboarding
        new.id = 0
        new.name = name
        new.options = options
        new.required = required
        new.single_select = single_select
        new.type = prompt_type
        return new
    
    
    def iter_options(self):
        """
        Iterates over the options of the prompt.
        
        This method is an iterable generator.
        
        Yields
        ------
        option : ``OnboardingPromptOption``
        """
        options = self.options
        if (options is not None):
            yield from options



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, nullable_date_time_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    nullable_date_time_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_array_validator_factory, entity_id_validator_factory,
    nullable_date_time_validator_factory, preinstanced_validator_factory
)
from ...guild import Guild
from ...user import ClientUserBase

from ..application import Application
from ..sku import SKU
from ..subscription import Subscription

from .preinstanced import EntitlementOwnerType, EntitlementType


# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id = entity_id_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', Application)

# consumed

parse_consumed = bool_parser_factory('consumed', False)
put_consumed = bool_optional_putter_factory('consumed', False)
validate_consumed = bool_validator_factory('consumed', False)

# deleted

parse_deleted = bool_parser_factory('deleted', False)
put_deleted = bool_optional_putter_factory('deleted', False)
validate_deleted = bool_validator_factory('deleted', False)

# ends_at

parse_ends_at = nullable_date_time_parser_factory('ends_at')
put_ends_at = nullable_date_time_optional_putter_factory('ends_at')
validate_ends_at = nullable_date_time_validator_factory('ends_at')

# exclude_deleted

validate_exclude_deleted = bool_validator_factory('exclude_deleted', False)

# exclude_ended

validate_exclude_ended = bool_validator_factory('exclude_ended', False)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('entitlement_id')

# owner

def put_owner(owner, data, defaults):
    """
    Serialises the entitlement's owner's type into the given data.
    
    Parameters
    ----------
    owner : `(EntitlementOwnerType, int)`
        The entitlement's owner represented by an ``EntitlementOwnerType`` and its identifier.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    put_owner_type(owner[0], data, defaults)
    put_owner_id(owner[1], data, defaults)
    return data


def validate_owner(owner):
    """
    Validates an entitlement's owner.
    
    Parameters
    ----------
    owner : `None`, ``ClientUserBase``, ``Guild``, `(int | EntitlementOwnerType, int | str)`
        The owner to validate.
    
    Returns
    -------
    owner : `(EntitlementOwnerType, int)`
    
    Raises
    ------
    TypeError
        - If `owner`'s type is incorrect.
    ValueError
        - If `owner`'s value is incorrect.
    """
    if owner is None:
        return EntitlementOwnerType.none, 0
    
    if isinstance(owner, Guild):
        return EntitlementOwnerType.guild, owner.id
    
    if isinstance(owner, ClientUserBase):
        return EntitlementOwnerType.user, owner.id
    
    if isinstance(owner, tuple):
        if len(owner) == 2:
            return validate_owner_type(owner[0]), validate_owner_id(owner[1])
        
        raise ValueError(
            f'When `owner` is given as a `tuple` its length must be `2`, got {len(owner)!r}; owner = {owner!r}.'
        )
    
    raise TypeError(
        f'`owner` can be `None`, `{ClientUserBase.__name__}`, `{Guild.__name__}` or a `tuple` of'
        f'`{EntitlementOwnerType.__name__}` and `int` representing it. Got {type(owner).__name__}; {owner!r}.'
    )


# owner_id

put_owner_id = entity_id_putter_factory('owner_id')
validate_owner_id = entity_id_validator_factory('owner_id')

# owner_type

put_owner_type = preinstanced_putter_factory('owner_type')
validate_owner_type = preinstanced_validator_factory('owner_type', EntitlementOwnerType)

# sku_id

parse_sku_id = entity_id_parser_factory('sku_id')
put_sku_id = entity_id_putter_factory('sku_id')
validate_sku_id = entity_id_validator_factory('sku_id', SKU)

# sku_ids

validate_sku_ids = entity_id_array_validator_factory('sku_ids', SKU)

# starts_at

parse_starts_at = nullable_date_time_parser_factory('starts_at')
put_starts_at = nullable_date_time_optional_putter_factory('starts_at')
validate_starts_at = nullable_date_time_validator_factory('starts_at')

# subscription_id

parse_subscription_id = entity_id_parser_factory('subscription_id')
put_subscription_id = entity_id_optional_putter_factory('subscription_id')
validate_subscription_id = entity_id_validator_factory('subscription_id', Subscription)

# type

parse_type = preinstanced_parser_factory('type', EntitlementType, EntitlementType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('entitlement_type', EntitlementType)

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_optional_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ()

from ...field_parsers import (
    nullable_date_time_parser_factory, nullable_object_array_parser_factory, nullable_string_parser_factory
)
from ...field_putters import (
    force_bool_putter_factory, nullable_date_time_optional_putter_factory,
    nullable_object_array_optional_putter_factory, nullable_string_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, nullable_date_time_validator_factory, nullable_object_array_validator_factory,
    nullable_string_validator_factory
)

from ..verification_screen_step import VerificationScreenStep

from .constants import DESCRIPTION_LENGTH_MAX

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_optional_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# edited_at

parse_edited_at = nullable_date_time_parser_factory('version')
put_edited_at = nullable_date_time_optional_putter_factory('version')
validate_edited_at = nullable_date_time_validator_factory('edited_at')

# enabled

put_enabled = force_bool_putter_factory('enabled')
validate_enabled = bool_validator_factory('enabled', True)

# steps

parse_steps = nullable_object_array_parser_factory('form_fields', VerificationScreenStep)
put_steps = nullable_object_array_optional_putter_factory('form_fields')
validate_steps = nullable_object_array_validator_factory('steps', VerificationScreenStep)



__all__ = ()

from warnings import warn

from scarletio import Compound

from ...core import GUILDS
from ...emoji import Emoji
from ...emoji.emoji.utils import EMOJI_FIELD_CONVERTERS
from ...http import DiscordApiClient
from ...payload_building import build_create_payload, build_edit_payload
from ...utils import image_to_base64

from ..request_helpers import get_emoji_and_guild_id_and_id, get_emoji_guild_id_and_id, get_guild_and_id, get_guild_id


class ClientCompoundEmojiGuildEndpoints(Compound):
    
    api : DiscordApiClient
    
    async def emoji_get_guild(self, emoji, *, force_update = False):
        """
        Requests the emoji by it's id at the given guild.
        If the client's logging in is finished, then it should have all of its emojis loaded already.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `(int, int)`
            The emoji, or 2 snowflake representing it.
        
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the emoji should be requested even if it supposed to be up to date.
        
        Returns
        -------
        emoji : ``Emoji``
        
        Raises
        ------
        TypeError
            - If `emoji` is given as incorrect type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        emoji, guild_id, emoji_id = get_emoji_and_guild_id_and_id(emoji)
        
        # If the emoji has no linked guild, we cannot request it, so we return instantly.
        if not guild_id:
            # Create empty emoji instead of returning `None` to stop expected errors
            if emoji is None:
                emoji = Emoji.precreate(emoji_id)
            
            return emoji
        
        if (not force_update) and (emoji is not None) and (not emoji.partial):
            return emoji
        
        emoji_data = await self.api.emoji_get_guild(guild_id, emoji_id)
        
        if (emoji is None):
            emoji = Emoji.from_data(emoji_data, guild_id)
        else:
            emoji._set_attributes(emoji_data, guild_id)
        
        return emoji
    
    
    async def emoji_get(self, emoji, *, force_update = False):
        """
        Deprecated and will be removed in 2025 January.
        """
        warn(
            (
                f'`{type(self).__name__}.emoji_get` is deprecated and will be removed in 2025 January.'
                f' Please use `.emoji_get_guild` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.emoji_get_guild(emoji, force_update = force_update)
    
    
    async def emoji_get_all_guild(self, guild):
        """
        Requests the given guild's emojis.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's emojis will be synced.
        
        Returns
        -------
        emojis : `list` of ``Emoji``
            The guild's emojis.
        
        Raises
        ------
        TypeError
            - If `guild`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        emoji_datas = await self.api.emoji_get_all_guild(guild_id)
        
        if guild is None:
            # Do not create a partial guild, because it would have been garbage collected after leaving the function
            # anyways
            guild = GUILDS.get(guild_id, None)
        
        if guild is None:
            emojis = [Emoji.from_data(emoji_data, guild_id) for emoji_data in emoji_datas]
        else:
            guild._update_emojis(emoji_datas)
            emojis = [*guild.iter_emojis()]
        
        return emojis
    
    
    async def emoji_guild_get_all(self, guild):
        """
        Deprecated and will be removed in 2025 January.
        """
        warn(
            (
                f'`{type(self).__name__}.emoji_guild_get_all` is deprecated and will be removed in 2025 January.'
                f' Please use `.emoji_get_all_guild` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.emoji_get_all_guild(guild)
    
    
    async def emoji_create_guild(self, guild, image, emoji_template = None, *, reason = None, **keyword_parameters):
        """
        Creates an emoji at the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild where the emoji will be created.
        
        image : `bytes-like`
            The emoji's icon. Up to 256 kB.
        
        emoji_template : `None`, ``Emoji`` = `None`, Optional
            Emoji entity to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the `guild`'s audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the emoji with.
        
        Other attributes
        ----------------
        name : `str`, Optional (Keyword only)
            The emoji's name. It's length can be between `2` and `32`.
        
        role_ids : `None | iterable<int> | iterable<Role>` = `None`, Optional (Keyword only)
            Whether the created emoji should be limited only to users with any of the specified roles.
        
        roles : `None | iterable<int> | iterable<Role>` = `None`, Optional (Keyword only)
            Alias for `role_ids`.
        
        Returns
        -------
        emoji : ``Emoji``
            The created emoji.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Only some characters can be in the emoji's name, so every other character is filtered out.
        """
        if isinstance(image, str) and isinstance(emoji_template, (bytes, bytearray, memoryview)):
            warn(
                (
                    f'`{type(self).__name__}.emoji_create`\'s `name` parameter is now keyword only. '
                    f'Please check the new definition for more details. '
                    f'Passing the name parameter as positional is deprecated and will be removed in 2025 January.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            keyword_parameters['name'] = image
            image = emoji_template
            emoji_template = None
        
        
        guild_id = get_guild_id(guild)
        data = build_create_payload(emoji_template, EMOJI_FIELD_CONVERTERS, keyword_parameters)
        data['image'] = image_to_base64(image)
        emoji_data = await self.api.emoji_create_guild(guild_id, data, reason)
        return Emoji.from_data(emoji_data, guild_id)
    
    
    async def emoji_edit_guild(self, emoji, emoji_template = None, *, reason = None, **keyword_parameters):
        """
        Edits the given (guild) emoji.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `(int, int)`
            The emoji to edit.
        
        emoji_template : `None`, ``Emoji`` = `None`, Optional
            A emoji to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Other attributes
        ----------------
        name : `str`, Optional (Keyword only)
            The emoji's name. It's length can be between `2` and `32`.
        
        role_ids : `None | iterable<int> | iterable<Role>` = `None`, Optional (Keyword only)
            Whether the created emoji should be limited only to users with any of the specified roles.
        
        roles : `None | iterable<int> | iterable<Role>` = `None`, Optional (Keyword only)
            Alias for `role_ids`.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        emoji, guild_id, emoji_id = get_emoji_and_guild_id_and_id(emoji)
        
        # Cannot edit partially loaded emojis.
        if not guild_id:
            return
        
        data = build_edit_payload(emoji, emoji_template, EMOJI_FIELD_CONVERTERS, keyword_parameters)
        
        if data:
            await self.api.emoji_edit_guild(guild_id, emoji_id, data, reason)
    
    
    async def emoji_delete_guild(self, emoji, *, reason = None):
        """
        Deletes the given emoji.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `(int, int)`
            The emoji to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If emoji is given as incorrect type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, emoji_id = get_emoji_guild_id_and_id(emoji)
        
        # Cannot delete partially loaded emoji.
        if not guild_id:
            return
        
        await self.api.emoji_delete_guild(guild_id, emoji_id, reason)
    
    
    emoji_create = emoji_create_guild
    emoji_edit = emoji_edit_guild
    emoji_delete = emoji_delete_guild



__all__ = ()


def _split_part_by_numbers(part):
    """
    Splits the given string part by numbers. Used by ``_split_bot_name_to_words``.
    
    This function is an iterable generator.
    
    Examples
    --------
    ```py
    325 -> 325
    koishi -> koishi
    koishi3 -> koishi, 3
    3koishi -> 3, koishi
    46koishi12satori -> 46, satori, 12, satori
    ```
    
    Parameters
    ----------
    part : `str`
        String part.
    
    Yields
    ------
    part : `str`
    """
    previous_characters = []
    
    for character in part:
        if previous_characters and previous_characters[-1].isnumeric() != character.isnumeric():
            yield ''.join(previous_characters)
            previous_characters.clear()
        
        previous_characters.append(character)
        continue
    
    if previous_characters:
        yield ''.join(previous_characters)


def _split_part_by_casing(part):
    """
    Splits the given string parts by casing. Used by ``_split_bot_name_to_words``.
    
    This function is an iterable generator.
    
    Examples
    --------
    ```py
    koishi -> koishi
    KOISHI -> KOISHI
    Koishi -> Koishi
    koishiSatori -> koishi, Satori
    KOISHISatori -> KOISHI, Satori
    koishiSATORI -> koishi, SATORI
    KOISHIsatori -> KOISH, Isatori # There is a limit how much we can do
    ```
    
    Parameters
    ----------
    part : `str`
        String part.
    
    Yields
    ------
    part : `str`
    """
    previous_characters = []
    
    for character in part:
        if previous_characters:
            if character.isupper():
                if not previous_characters[-1].isupper():
                    yield ''.join(previous_characters)
                    previous_characters.clear()
            
            else:
                if previous_characters[-1].isupper() and len(previous_characters) > 1:
                    yield ''.join(previous_characters[:-1])
                    del previous_characters[:-1]
        
        previous_characters.append(character)
        continue
    
    if previous_characters:
        yield ''.join(previous_characters)


def _split_bot_name_to_words(bot_name):
    """
    Splits the given name to words. Used by ``get_bot_module_name``, ``get_bot_variable_name`` and
    ``get_bot_constant_name``.
    
    This function is an iterable generator.
    
    Examples
    --------
    ```py
    koishi_satori_69 -> koishi, satori, 69
    KoishiSatori69 -> Koishi, Satori, 69
    KOISHI_SATORI_69 -> KOISHI, Satori, 69
    KOISHISATORI69 -> KOISHISATORI, 69 # there is a limit how much we can do
    ```
    
    Parameters
    ----------
    bot_name : `str`
        A bot's name.
    
    Yields
    ------
    word : `str`
    """
    for part in bot_name.split('_'):
        if part:
            for part in _split_part_by_numbers(part):
                yield from _split_part_by_casing(part)


def get_bot_module_name(bot_name):
    """
    Creates bot module name.
    
    Parameters
    ----------
    bot_name : `str`
        A bot's name.
    
    Returns
    -------
    module_name : `str`
    """
    return '_'.join([part.lower() for part in _split_bot_name_to_words(bot_name)])
    

def get_bot_variable_name(bot_name):
    """
    Creates bot variable name.
    
    Parameters
    ----------
    bot_name : `str`
        A bot's name.
    
    Returns
    -------
    variable_name : `str`
    """
    return ''.join([part.capitalize() for part in _split_bot_name_to_words(bot_name)])


def get_bot_constant_name(bot_name):
    """
    Creates bot constant name.
    
    Parameters
    ----------
    bot_name : `str`
        A bot's name.
    
    Returns
    -------
    constant_name : `str`
    """
    return '_'.join([part.upper() for part in _split_bot_name_to_words(bot_name)])


def get_bot_display_name(bot_name):
    """
    Creates the bot's display name.
    
    Parameters
    ----------
    bot_name : `str`
        A bot's name.
    
    Returns
    -------
    display_name : `str`
    """
    return ' '.join([part.capitalize() for part in _split_bot_name_to_words(bot_name)])


def get_project_module_name(project_name):
    """
    Gets the project's name.
    
    Parameters
    ----------
    project_name : `str`
        The project's name.
    
    Returns
    -------
    project_module_name : `str`
    """
    return get_bot_module_name(project_name)



__all__ = (
    'FailingAddress', 'RoutePlannerBase', 'RoutePlannerNanoIP', 'RoutePlannerRotatingIp', 'RoutePlannerRotatingNanoIP'
)

import warnings

from scarletio import RichAttributeErrorBaseType, copy_docs

from ...discord.utils import DATETIME_FORMAT_CODE, unix_time_to_datetime

from .constants import (
    LAVALINK_KEY_ROUTEPLANNER_OBJECT, LAVALINK_KEY_ROUTEPLANNER_OBJECT_BLOCK_INDEX_STRING,
    LAVALINK_KEY_ROUTEPLANNER_OBJECT_CURRENT_ADDRESS, LAVALINK_KEY_ROUTEPLANNER_OBJECT_CURRENT_ADDRESS_INDEX_STRING,
    LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESSES, LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESS_ADDRESS,
    LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESS_UNIX_TIME, LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK,
    LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK_SIZE_STRING, LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK_TYPE,
    LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_INDEX_STRING, LAVALINK_KEY_ROUTEPLANNER_OBJECT_ROTATE_INDEX_STRING,
    LAVALINK_KEY_ROUTEPLANNER_TYPE, LAVALINK_KEY_ROUTEPLANNER_TYPE_NANO_IP, LAVALINK_KEY_ROUTEPLANNER_TYPE_ROTATING_IP,
    LAVALINK_KEY_ROUTEPLANNER_TYPE_ROTATING_NANO_IP
)


class FailingAddress(RichAttributeErrorBaseType):
    """
    Failed address.
    
    Attributes
    ----------
    address : `str`
        The failed address.
    time : `datetime`
        The time, when the fail occurred.
    """
    __slots__ = ('address', 'time')
    
    def __new__(cls, data):
        """
        Creates a new failing address instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Failing address data.
        """
        self = object.__new__(cls)
        self.address = data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESS_ADDRESS]
        self.time = unix_time_to_datetime(data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESS_UNIX_TIME])
        return self
    
    def __repr__(self):
        """Returns the failing address's representation."""
        return f'<{self.__class__.__name__} address={self.address!r}, time={self.time:{DATETIME_FORMAT_CODE}}>'


def _route_planner_base_repr_generator(route_planner):
    """
    Representation builder for ``RoutePlannerBase``-s.
    
    Parameters
    ----------
    route_planner : ``RoutePlannerBase``.
        The instance to get representation of.
    
    Yields
    ------
    repr_parts : `list` of `str`
        Repr parts to extend.
    
    Examples
    --------
    
    ```py
    for repr_parts in _route_planner_base_repr_generator(route_planner):
        repr_parts.append('...')
    
    return ''.join(repr_parts)
    ```
    """
    repr_parts = ['<', route_planner.__class__.__name__]
    
    repr_parts.append(' ip_block_type = ')
    repr_parts.append(route_planner.ip_block_type)
    
    repr_parts.append(', ip_block_size = ')
    repr_parts.append(repr(route_planner.ip_block_size))
    
    yield repr_parts
    
    failing_addresses = route_planner.failing_addresses
    if (failing_addresses is None):
        repr_parts.append(', failing_addresses = [')
        
        length = len(failing_addresses)
        index = 0
        
        while True:
            failing_address = failing_addresses[index]
            repr_parts.append(repr(failing_address))
            
            index += 1
            if index == length:
                break
            
            repr_parts.append(', ')
            continue
        
        repr_parts.append(']')
    
    repr_parts.append('>')


class RoutePlannerBase(RichAttributeErrorBaseType):
    """
    Base class of route planners.
    
    Attributes
    ----------
    failing_addresses : `None`, `tuple` of ``FailingAddress``
        The failed addresses.
    ip_block_size : `int`
        The ip block's size.
    ip_block_type : `str`
        The ip block's type.
    """
    __slots__ = 'failing_addresses', 'ip_block_size', 'ip_block_type'
    
    def __new__(cls, data):
        """
        Creates a new route planner instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Route planner data.
        """
        self = object.__new__(cls)
        
        ip_block_data = data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK]
        self.ip_block_type = ip_block_data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK_TYPE]
        self.ip_block_size = int(ip_block_data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK_SIZE_STRING])
        
        failing_address_datas = data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESSES]
        if failing_address_datas:
            failing_addresses = tuple(
                FailingAddress(failing_address_data) for failing_address_data in failing_address_datas
            )
        else:
            failing_addresses = None
        self.failing_addresses = failing_addresses
        
        return self
    
    def __repr__(self):
        """Returns the route planner's representation."""
        for repr_parts in _route_planner_base_repr_generator(self):
            pass
        
        return ''.join(repr_parts)


class RoutePlannerRotatingIp(RoutePlannerBase):
    """
    Ip rotating route planner.
    
    Attributes
    ----------
    failing_addresses : `None`, `tuple` of ``FailingAddress``
        The failed addresses.
    ip_block_size : `int`
        The ip block's size.
    ip_block_type : `str`
        The ip block's type.
    current_address : `str`
        The currently used ip address.
    ip_index : `int`
        The current offset in the block.
    rotate_index : `int`
        The number of rotations, since the lavalink server was restarted.
    """
    __slots__ = ('current_address', 'ip_index', 'rotate_index')
    
    @copy_docs(RoutePlannerBase.__new__)
    def __new__(cls, data):
        self = RoutePlannerBase.__new__(cls, data)
        self.rotate_index = int(data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_ROTATE_INDEX_STRING])
        self.ip_index = int(data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_INDEX_STRING])
        self.current_address = data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_CURRENT_ADDRESS]
        return self
    
    @copy_docs(RoutePlannerBase.__repr__)
    def __repr__(self):
        for repr_parts in _route_planner_base_repr_generator(self):
            repr_parts.append(', rotate_index = ')
            repr_parts.append(repr(self.rotate_index))
            
            repr_parts.append(', ip_index = ')
            repr_parts.append(repr(self.ip_index))
            
            repr_parts.append(', current_address = ')
            repr_parts.append(repr(self.current_address))
        
        return ''.join(repr_parts)


class RoutePlannerNanoIP(RoutePlannerBase):
    """
    Switches the IP on each clock update and uses the current nanosecond in the offset in the used block. This strategy
    requires at least a (combined) /64 IPv6 block (2⁶⁴ addresses).
    
    Attributes
    ----------
    failing_addresses : `None`, `tuple` of ``FailingAddress``
        The failed addresses.
    ip_block_size : `int`
        The ip block's size.
    ip_block_type : `str`
        The ip block's type.
    current_address_index : `int`
        Representing the current offset in the ip block
    """
    __slots__ = ('current_address_index', )
    
    @copy_docs(RoutePlannerBase.__new__)
    def __new__(cls, data):
        self = RoutePlannerBase.__new__(cls, data)
        self.current_address_index = int(data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_CURRENT_ADDRESS_INDEX_STRING])
        return self
    
    
    @copy_docs(RoutePlannerBase.__repr__)
    def __repr__(self):
        for repr_parts in _route_planner_base_repr_generator(self):
            repr_parts.append(', current_address_index = ')
            repr_parts.append(repr(self.current_address_index))
        
        return ''.join(repr_parts)


class RoutePlannerRotatingNanoIP(RoutePlannerNanoIP):
    """
    Switches the IP on each clock update and uses the current nanosecond in the offset of the used block.

    When a ban occurs, this strategy rotates to the next /64 block as fallback strategy. This strategy requires at
    least a /64 IPv6 CIDR (2⁶⁴ addresses).
    
    Attributes
    ----------
    failing_addresses : `None`, `tuple` of ``FailingAddress``
        The failed addresses.
    ip_block_size : `int`
        The ip block's size.
    ip_block_type : `str`
        The ip block's type.
    current_address_index : `int`
        Representing the current offset in the ip block
    block_index : `int`
        Contains which /64 block ips are chosen. This number increases on each ban.
    """
    __slots__ = ('block_index', )
    
    @copy_docs(RoutePlannerBase.__new__)
    def __new__(cls, data):
        self = RoutePlannerNanoIP.__new__(cls, data)
        self.block_index = int(data[LAVALINK_KEY_ROUTEPLANNER_OBJECT_BLOCK_INDEX_STRING])
        return self
    
    @copy_docs(RoutePlannerBase.__repr__)
    def __repr__(self):
        for repr_parts in _route_planner_base_repr_generator(self):
            repr_parts.append(', current_address_index = ')
            repr_parts.append(repr(self.current_address_index))
            
            repr_parts.append(', block_index = ')
            repr_parts.append(repr(self.block_index))
            
        return ''.join(repr_parts)

ROUTE_PLANNER_TYPES = {
    LAVALINK_KEY_ROUTEPLANNER_TYPE_ROTATING_IP : RoutePlannerRotatingIp,
    LAVALINK_KEY_ROUTEPLANNER_TYPE_NANO_IP : RoutePlannerNanoIP,
    LAVALINK_KEY_ROUTEPLANNER_TYPE_ROTATING_NANO_IP : RoutePlannerRotatingNanoIP,
}

def get_route_planner(data):
    """
    Tries to get the routeplanner based on it's type.
    
    Parameters
    ----------
    data : `dict` Of (`str`, `object`) items
        Route planner data.
    
    Returns
    -------
    route_planner : `None`, ``RoutePlannerBase``
    """
    route_planner_type_name = data[LAVALINK_KEY_ROUTEPLANNER_TYPE]
    if route_planner_type_name is None:
        route_planner = None
    else:
        try:
            route_planner_type = ROUTE_PLANNER_TYPES[route_planner_type_name]
        except KeyError:
            warnings.warn(
                f'Undefined route planner type : {route_planner_type_name!r}\nPlease open an issue with this message.',
                RuntimeWarning,
            )
            
            route_planner = None
        
        else:
            route_planner_object_data = data[LAVALINK_KEY_ROUTEPLANNER_OBJECT]
            route_planner = route_planner_type(route_planner_object_data)
    
    return route_planner



__all__ = ('IntegrationType', )

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase

from ..integration_metadata import IntegrationMetadataBase, IntegrationMetadataSubscription, IntegrationMetadataDiscord


@export
class IntegrationType(PreinstancedBase, value_type = str):
    """
    Represents an ``Integration``'s type.
    
    Attributes
    ----------
    metadata_type : `type<IntegrationMetadataBase>`
        The integration's metadata's type.
        
    name : `str`
        The name of the integration type.
    
    value : `str`
        The Discord side identifier value of the integration type.
    
    Type Attributes
    ---------------
    Every predefined integration type can be accessed as type attribute as well:
    
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | Type attribute name   | Name                  | Value                     | Metadata type                         |
    +=======================+=======================+===========================+=======================================+
    | none                  | none                  | `''`                      | ``IntegrationMetadataBase``           |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | discord               | Discord               | `'discord'`               | ``IntegrationMetadataDiscord``        |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | twitch                | Twitch                | `'twitch'`                | ``IntegrationMetadataSubscription``   |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | youtube               | Youtube               | `'youtube'`               | ``IntegrationMetadataSubscription``   |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | guild_subscription    | guild subscription    | `'guild_subscription'`    | ``IntegrationMetadataSubscription``   |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    def __new__(cls, value, name = None, metadata_type = None):
        """
        Creates a new integration type.
        
        Parameters
        ----------
        value : `str`
            The unique identifier of the scheduled event entity type.
        
        name : `None | str` = `None`, Optional
            The name of the scheduled event entity type.
        
        metadata_type : `None | type<IntegrationMetadataBase>` = `None`, Optional
            The integration's metadata's type.
        """
        if metadata_type is None:
            metadata_type = IntegrationMetadataBase
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.metadata_type = metadata_type
        return self
    
    
    # predefined
    none = P('', 'none', IntegrationMetadataBase)
    discord = P('discord', 'Discord', IntegrationMetadataDiscord)
    twitch = P('twitch', 'twitch', IntegrationMetadataSubscription)
    youtube = P('youtube', 'Youtube', IntegrationMetadataSubscription)
    guild_subscription = P('guild_subscription', 'guild subscription', IntegrationMetadataSubscription)



__all__ = ('FlagDescriptor',)

from scarletio import RichAttributeErrorBaseType, copy_docs


NAME_UNKNOWN = '<unknown>'


class FlagDescriptorBase(RichAttributeErrorBaseType):
    """
    Base flag descriptor.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Raises `NotImplementedError`.
        """
        raise NotImplementedError
    
    
    def _build_cannot_exception(self, action):
        """
        Builds a "cannot" exception to be raised.
        
        Parameters
        ----------
        action : `str`
            The action that cannot be executed.
        
        Returns
        -------
        exception : `AttributeError`
        """
        return AttributeError(f'Cannot {action} `{self.type_name}.{self.flag_name}`.')
    
    
    def __get__(self, instance, instance_type):
        """
        Gets the field with the conversion.
        If accessed as a type attribute returns itself.
        
        Parameters
        ----------
        instance : `None | FlagBase`
            Instance to set the fields to.
        
        instance_type : ``FlagMeta``
            The instance's type.
        
        Returns
        -------
        value : `bool | self`
        """
        if instance is None:
            return self
        
        return False
    
    
    def __set__(self, instance, value):
        """
        Sets the field.
        
        Parameters
        ----------
        instance : `FlagBase`
            Instance to set the fields to.
        
        value : `object`
            The value to set.
        
        Raises
        ------
        AttributeError
        """
        raise self._build_cannot_exception('set')
    
    
    def __delete__(self, instance):
        """
        Deletes the field.
        
        Parameters
        ----------
        instance : `FlagBase`
            Instance to set the fields to.
        
        Raises
        ------
        AttributeError
        """
        raise self._build_cannot_exception('delete')
    
    
    def __repr__(self):
        """Returns the flag descriptor's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # shift
        repr_parts.append(', shift = ')
        repr_parts.append(repr(self.shift))
        
        # deprecation
        deprecation = self.deprecation
        if (deprecation is not None):
            repr_parts.append(', deprecation = ')
            repr_parts.append(repr(deprecation))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two flag descriptors are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # mask
        if self.mask != other.mask:
            return False
        
        # flag_name
        if self.flag_name != other.flag_name:
            return False
        
        # shift
        if self.shift != other.shift:
            return False
        
        # type_name
        if self.type_name != other.type_name:
            return False
        
        # deprecation
        if self.deprecation != other.deprecation:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the flag descriptor's hash value."""
        hash_value = 0
        
        # mask
        hash_value ^= self.mask
        
        # flag_name
        hash_value ^= hash(self.flag_name)
        
        # shift
        hash_value ^= hash(self.shift)
        
        # type_name
        hash_value ^= hash(self.type_name)
        
        # deprecation
        deprecation = self.deprecation
        if (deprecation is not None):
            hash_value ^= hash(deprecation)
        
        return hash_value
    
    
    @property
    def flag_name(self):
        """
        Returns the flag's name.
        
        Returns
        -------
        flag_name : `str`
        """
        return NAME_UNKNOWN
    
    
    @property
    def type_name(self):
        """
        Returns the flag's type's name.
        
        Returns
        -------
        type_name : `str`
        """
        return NAME_UNKNOWN
    
    
    @property
    def mask(self):
        """
        Returns the flag's mask.
        
        Returns
        -------
        mask : `int`
        """
        return 0
    
    
    @property
    def shift(self):
        """
        Returns the flag's bit shift.
        
        Returns
        -------
        shift : `int`
        """
        return 0
    
    
    @property
    def name(self):
        """
        Returns the flags full name.
        
        Returns
        -------
        name : `str`
        """
        return f'{self.type_name}.{self.flag_name}'
    
    
    @property
    def deprecation(self):
        """
        Returns deprecation notice of this flag if available.
        
        Returns
        -------
        deprecation : `None | FlagDeprecation`
        """
        return None


class FlagDescriptor(FlagDescriptorBase):
    """
    A simple descriptors that can be assigned at type body and translated to an actually functioning descriptor at
    creation.
    
    Attributes
    ----------
    shift : `int`
        The represented bit by its shift.
    
    deprecation : `None | FlagDeprecation`
        Deprecation notice for this flag.
    """
    __slots__ = ('deprecation', 'shift',)
    
    def __new__(cls, shift, *, deprecation = None):
        """
        Creates a new flag descriptor.
        
        Parameters
        ----------
        shift : `int`
            The represented bit by its shift.
        
        deprecation : `None | FlagDeprecation` = `None`, Optional (Keyword only)
            Deprecation notice for this flag.
        """
        # Do not set deprecation if not allowed.
        if (deprecation is not None) and (not deprecation.allowed):
            deprecation = None
        
        self = object.__new__(cls)
        self.deprecation = deprecation
        self.shift = shift
        return self
    
    
    @property
    @copy_docs(FlagDescriptorBase.mask)
    def mask(self):
        return 1 << self.shift


class FlagBitDescriptor(FlagDescriptorBase):
    """
    Flag bit descriptor.
    
    Attributes
    ----------
    flag_name : `str`
        The flag's name.
    
    mask : `int`
        Bit mask created from ``.shift``.
    
    shift : `int`
        The represented bit by its shift.
    
    type_name : `str`
        The type's name the descriptor is under.
    """
    __slots__ = ('flag_name', 'mask', 'shift', 'type_name')
    
    def __new__(cls, shift, type_name, flag_name):
        """
        Creates a new bit descriptor.
        
        Attributes
        ----------
        flag_name : `str`
            The flag's name.
        
        shift : `int`
            The represented bit by its shift.
        
        type_name : `str`
            The type's name the descriptor is under.
        """
        self = object.__new__(cls)
        self.flag_name = flag_name
        self.mask = 1 << shift
        self.shift = shift
        self.type_name = type_name
        return self
    
    
    @copy_docs(FlagDescriptorBase.__get__)
    def __get__(self, instance, instance_type):
        if instance is None:
            return self
        
        return True if instance & self.mask else False


class FlagBitDescriptorDeprecated(FlagBitDescriptor):
    """
    Flag bit descriptor.
    
    Attributes
    ----------
    deprecation : `FlagDeprecation`
        Deprecation notice for this flag.
    
    flag_name : `str`
        The flag's name.
    
    mask : `int`
        Bit mask created from ``.shift``.
    
    shift : `int`
        The represented bit by its shift.
    
    type_name : `str`
        The type's name the descriptor is under.
    """
    __slots__ = ('deprecation',)
    
    def __new__(cls, shift, type_name, flag_name, deprecation):
        """
        Creates a new bit descriptor.
        
        Attributes
        ----------
        flag_name : `str`
            The flag's name.
        
        shift : `int`
            The represented bit by its shift.
        
        type_name : `str`
            The type's name the descriptor is under.
        
        deprecation : `FlagDeprecation`
            Deprecation notice for this flag.
        """
        self = FlagBitDescriptor.__new__(cls, shift, type_name, flag_name)
        self.deprecation = deprecation
        return self
    
    
    @copy_docs(FlagDescriptorBase.__get__)
    def __get__(self, instance, instance_type):
        if instance is None:
            return self
        
        self.deprecation.trigger(self.type_name, self.flag_name, 3)
        return True if instance & self.mask else False


class FlagBitDescriptorReversed(FlagBitDescriptor):
    """
    Flag bit descriptor that reverses its return value.
    
    Attributes
    ----------
    flag_name : `str`
        The flag's name.
    
    mask : `int`
        Bit mask created from ``.shift``.
    
    shift : `int`
        The represented bit by its shift.
    
    type_name : `str`
        The type's name the descriptor is under.
    """
    __slots__ = ()
    
    @copy_docs(FlagDescriptorBase.__get__)
    def __get__(self, instance, instance_type):
        if instance is None:
            return self
        
        return False if instance & self.mask else True


class FlagBitDescriptorReversedDeprecated(FlagBitDescriptorDeprecated, FlagBitDescriptorReversed):
    """
    Flag bit descriptor that reverses its return value.
    
    Attributes
    ----------
    deprecation : `FlagDeprecation`
        Deprecation notice for this flag.
    
    flag_name : `str`
        The flag's name.
    
    mask : `int`
        Bit mask created from ``.shift``.
    
    shift : `int`
        The represented bit by its shift.
    
    type_name : `str`
        The type's name the descriptor is under.
    """
    __slots__ = ()
    
    @copy_docs(FlagDescriptorBase.__get__)
    def __get__(self, instance, instance_type):
        if instance is None:
            return self
        
        self.deprecation.trigger(self.type_name, self.flag_name, 3)
        return False if instance & self.mask else True



__all__ = ('Schedule',)

from datetime import datetime as DateTime

from scarletio import RichAttributeErrorBaseType

from ...utils import DATETIME_FORMAT_CODE, DISCORD_EPOCH_START

from ..schedule_nth_weeks_day import ScheduleNthWeeksDay, ScheduleWeeksDay

from .fields import (
    parse_by_month_days, parse_by_months, parse_by_nth_weeks_days, parse_by_weeks_days, parse_by_year_days, parse_end,
    parse_frequency, parse_occurrence_count_limit, parse_occurrence_spacing, parse_start, put_by_month_days,
    put_by_months, put_by_nth_weeks_days, put_by_weeks_days, put_by_year_days, put_end,
    put_frequency, put_occurrence_count_limit, put_occurrence_spacing, put_start,
    validate_by_month_days, validate_by_months, validate_by_nth_weeks_days, validate_by_weeks_days,
    validate_by_year_days, validate_end, validate_frequency, validate_occurrence_count_limit,
    validate_occurrence_spacing, validate_start
)
from .preinstanced import ScheduleFrequency, ScheduleMonth


class Schedule(RichAttributeErrorBaseType):
    """
    The schedule of a scheduled event.
    
    There are many limitation and many fields are also mutually exclusive.
    These are not checked directly here, so you will have tinker how things work by yourself.
    If you do not want to tinker, there are presents you may want to try out:
    - ``.create_weekly`` (example: every day )
    - ``.create_weekly`` (example: every sunday we go shopping)
    - ``.create_bi_weekly`` (example: every second saturday we visit people)
    - ``.create_monthly_nth_weeks_day`` (example: every month 1st week sunday we play chicken horse)
    - ``.create_yearly_month_nth_day`` (example: every year january 13 is someone's birthday)
    
    More may be added as required.
    
    Attributes
    ----------
    by_month_days : `None | tuple<int>`
        On which days of the month should the event occur at.
    
    by_months : `None | tuple<ScheduleMonth>`
        On which months should the event occur at.
    
    by_nth_weeks_days : `None | tuple<ScheduleNthWeeksDay>`
        On which days the event should occur at.
    
    by_weeks_days : `None | tuple<ScheduleWeeksDay>`
        In which days of the week should the event occur at.
    
    by_year_days : `None | tuple<int>`
        On which days of the year should the even occur at.
    
    end : `None | DateTime`
        When the occurrence should end at.
    
    frequency : ``ScheduleFrequency``
        How often should the event occur.
    
    occurrence_count_limit : `int`
        Up to how much times should the event occur.
    
    occurrence_spacing : `int`
        The spacing between 2 events.
        If set as `1` means it should occur on every occasion.
        If set to `2` it will occur only on every other.
    
    start : `None | DateTime`
        When the occurrence should start at.
    """
    __slots__ = (
        'by_month_days', 'by_months', 'by_nth_weeks_days', 'by_weeks_days', 'by_year_days', 'end', 'frequency',
        'occurrence_count_limit', 'occurrence_spacing', 'start'
    )
    
    def __new__(
        cls,
        by_month_days = ...,
        by_months = ...,
        by_nth_weeks_days = ...,
        by_weeks_days = ...,
        by_year_days = ...,
        end = ...,
        frequency = ...,
        occurrence_count_limit = ...,
        occurrence_spacing = ...,
        start = ...,
    ):
        """
        Creates a schedule with the given fields.
        
        Parameters
        ----------
        by_month_days : `None | iterable<int>`, Optional (Keyword only)
            On which days of the month should the event occur at.
        
        by_months : `None | iterable<ScheduleMonth> | iterable<int>`, Optional (Keyword only)
            On which months should the event occur at.
        
        by_nth_weeks_days : `None | iterable<ScheduleNthWeeksDay>`, Optional (Keyword only)
            On which days the event should occur at.
        
        by_weeks_days : `None | iterable<ScheduleWeeksDay> | iterable<int>`, Optional (Keyword only)
            In which days of the week should the event occur at.
        
        by_year_days : `None | iterable<int>`, Optional (Keyword only)
            On which days of the year should the even occur at.
        
        end : `None | DateTime`, Optional (Keyword only)
            When the occurrence should end at.
        
        frequency : `ScheduleFrequency | int`, Optional (Keyword only)
            How often should the event occur.
        
        occurrence_count_limit : `int`, Optional (Keyword only)
            Up to how much times should the event occur.
        
        occurrence_spacing : `int`, Optional (Keyword only)
            The spacing between 2 events.
        
        start : `None | DateTime`, Optional (Keyword only)
            When the occurrence should start at.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # by_month_days
        if by_month_days is ...:
            by_month_days = None
        else:
            by_month_days = validate_by_month_days(by_month_days)
        
        # by_months
        if by_months is ...:
            by_months = None
        else:
            by_months = validate_by_months(by_months)
        
        # by_nth_weeks_days
        if by_nth_weeks_days is ...:
            by_nth_weeks_days = None
        else:
            by_nth_weeks_days = validate_by_nth_weeks_days(by_nth_weeks_days)
        
        # by_weeks_days
        if by_weeks_days is ...:
            by_weeks_days = None
        else:
            by_weeks_days = validate_by_weeks_days(by_weeks_days)
        
        # by_year_days
        if by_year_days is ...:
            by_year_days = None
        else:
            by_year_days = validate_by_year_days(by_year_days)
        
        # end
        if end is ...:
            end = None
        else:
            end = validate_end(end)
        
        # frequency
        if frequency is ...:
            frequency = ScheduleFrequency.yearly
        else:
            frequency = validate_frequency(frequency)
        
        # occurrence_count_limit
        if occurrence_count_limit is ...:
            occurrence_count_limit = 0
        else:
            occurrence_count_limit = validate_occurrence_count_limit(occurrence_count_limit)
        
        # occurrence_spacing
        if occurrence_spacing is ...:
            occurrence_spacing = 1
        else:
            occurrence_spacing = validate_occurrence_spacing(occurrence_spacing)
        
        # start
        if start is ...:
            start = None
        else:
            start = validate_start(start)
        
        # construct
        self = object.__new__(cls)
        self.by_month_days = by_month_days
        self.by_months = by_months
        self.by_nth_weeks_days = by_nth_weeks_days
        self.by_weeks_days = by_weeks_days
        self.by_year_days = by_year_days
        self.end = end
        self.frequency = frequency
        self.occurrence_count_limit = occurrence_count_limit
        self.occurrence_spacing = occurrence_spacing
        self.start = start
        return self
    
    
    def __repr__(self):
        """Returns the representation of the schedule."""
        repr_parts = ['<', type(self).__name__]
        
        # frequency
        frequency = self.frequency
        repr_parts.append(' frequency = ')
        repr_parts.append(frequency.name)
        
        # yearlies
        if frequency is ScheduleFrequency.yearly:
            # by_year_days
            by_year_days = self.by_year_days
            if (by_year_days is not None):
                repr_parts.append(', by_year_days = ')
                repr_parts.append(repr(by_year_days))
            
            # by_months
            by_months = self.by_months
            if (by_months is not None):
                repr_parts.append(', by_months = ')
                repr_parts.append(repr(by_months))
        
        # yearlies and monthlies
        if frequency is ScheduleFrequency.yearly or ScheduleFrequency.monthly:
            # by_month_days
            by_month_days = self.by_month_days
            if (by_month_days is not None):
                repr_parts.append(', by_month_days = ')
                repr_parts.append(repr(by_month_days))
            
            # by_nth_weeks_days
            by_nth_weeks_days = self.by_nth_weeks_days
            if (by_nth_weeks_days is not None):
                repr_parts.append(', by_nth_weeks_days = ')
                repr_parts.append(repr(by_nth_weeks_days))
        
        # weeklies and dailies
        if frequency is ScheduleFrequency.weekly or frequency is ScheduleFrequency.daily:
            by_weeks_days = self.by_weeks_days
            if (by_weeks_days is not None):
                repr_parts.append(', by_weeks_days = ')
                repr_parts.append(repr(by_weeks_days))
        
        # occurrence_spacing
        occurrence_spacing = self.occurrence_spacing
        if (occurrence_spacing != 1):
            repr_parts.append(', occurrence_spacing = ')
            repr_parts.append(repr(occurrence_spacing))
        
        # occurrence_count_limit
        occurrence_count_limit = self.occurrence_count_limit
        if (occurrence_count_limit != 0):
            repr_parts.append(', occurrence_count_limit = ')
            repr_parts.append(repr(occurrence_count_limit))
        
        # start
        start = self.start
        if (start is not None):
            repr_parts.append(', start = ')
            repr_parts.append(format(start, DATETIME_FORMAT_CODE))
        
        # end
        end = self.end
        if (end is not None):
            repr_parts.append(', end = ')
            repr_parts.append(format(end, DATETIME_FORMAT_CODE))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two schedules are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.by_month_days != other.by_month_days:
            return False
        
        if self.by_months != other.by_months:
            return False
        
        if self.by_nth_weeks_days != other.by_nth_weeks_days:
            return False
        
        if self.by_weeks_days != other.by_weeks_days:
            return False
        
        if self.by_year_days != other.by_year_days:
            return False
        
        if self.end != other.end:
            return False
        
        if self.frequency is not other.frequency:
            return False
        
        if self.occurrence_count_limit != other.occurrence_count_limit:
            return False
        
        if self.occurrence_spacing != other.occurrence_spacing:
            return False
        
        if self.start != other.start:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the schedule's hash value."""
        hash_value = 0
        
        # by_month_days
        by_month_days = self.by_month_days
        if (by_month_days is not None):
            hash_value ^= len(by_month_days) << 0
            
            for by_month_day in by_month_days:
                hash_value ^= hash(by_month_day) << 1
        
        # by_months
        by_months = self.by_months
        if (by_months is not None):
            hash_value ^= len(by_months) << 2
            
            for by_month in by_months:
                hash_value ^= hash(by_month) << 3

        # by_nth_weeks_days
        by_nth_weeks_days = self.by_nth_weeks_days
        if (by_nth_weeks_days is not None):
            hash_value ^= len(by_nth_weeks_days) << 4
            
            for by_nth_weeks_day in by_nth_weeks_days:
                hash_value ^= hash(by_nth_weeks_day) << 5

        # by_weeks_days
        by_weeks_days = self.by_weeks_days
        if (by_weeks_days is not None):
            hash_value ^= len(by_weeks_days) << 6
            
            for by_weeks_day in by_weeks_days:
                hash_value ^= hash(by_weeks_day) << 7

        # by_year_days
        by_year_days = self.by_year_days
        if (by_year_days is not None):
            hash_value ^= len(by_year_days) << 8
            
            for by_year_day in by_year_days:
                hash_value ^= hash(by_year_day) << 9
        
        # end
        end = self.end
        if (end is not None):
            hash_value ^= hash(end)
        
        # frequency
        hash_value ^= hash(self.frequency) << 10
        
        # occurrence_count_limit
        hash_value ^= self.occurrence_count_limit << 11
        
        # occurrence_spacing
        hash_value ^= self.occurrence_spacing << 12
        
        # start
        start = self.start
        if (start is not None):
            hash_value ^= hash(start)
        
        return hash_value
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a schedule from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.by_month_days = parse_by_month_days(data)
        self.by_months = parse_by_months(data)
        self.by_nth_weeks_days = parse_by_nth_weeks_days(data)
        self.by_weeks_days = parse_by_weeks_days(data)
        self.by_year_days = parse_by_year_days(data)
        self.end = parse_end(data)
        self.frequency = parse_frequency(data)
        self.occurrence_count_limit = parse_occurrence_count_limit(data)
        self.occurrence_spacing = parse_occurrence_spacing(data)
        self.start = parse_start(data)
        return self
    
    
    def to_data(self, *, defaults = False, start = None):
        """
        Converts the schedule to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        start : `None | DateTime` = `None`, Optional (Keyword only)
            Start date to use if schedule does not define it.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_by_month_days(self.by_month_days, data, defaults)
        put_by_months(self.by_months, data, defaults)
        put_by_nth_weeks_days(self.by_nth_weeks_days, data, defaults)
        put_by_weeks_days(self.by_weeks_days, data, defaults)
        put_by_year_days(self.by_year_days, data, defaults)
        put_end(self.end, data, defaults)
        put_frequency(self.frequency, data, defaults)
        put_occurrence_count_limit(self.occurrence_count_limit, data, defaults)
        put_occurrence_spacing(self.occurrence_spacing, data, defaults)
        
        schedule_start = self.start
        if (schedule_start is None):
            if start is None:
                schedule_start = DISCORD_EPOCH_START
            else:
                schedule_start = start
        
        put_start(schedule_start, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the schedule.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # by_month_days
        by_month_days = self.by_month_days
        if (by_month_days is not None):
            by_month_days = (*by_month_days,)
        new.by_month_days = by_month_days
        
        # by_months
        by_months = self.by_months
        if (by_months is not None):
            by_months = (*by_months,)
        new.by_months = by_months
        
        # by_nth_weeks_days
        by_nth_weeks_days = self.by_nth_weeks_days
        if (by_nth_weeks_days is not None):
            by_nth_weeks_days = (*by_nth_weeks_days,)
        new.by_nth_weeks_days = by_nth_weeks_days
        
        # by_weeks_days
        by_weeks_days = self.by_weeks_days
        if (by_weeks_days is not None):
            by_weeks_days = (*by_weeks_days,)
        new.by_weeks_days = by_weeks_days
        
        # by_year_days
        by_year_days = self.by_year_days
        if (by_year_days is not None):
            by_year_days = (*by_year_days,)
        new.by_year_days = by_year_days
        
        new.end = self.end
        new.frequency = self.frequency
        new.occurrence_count_limit = self.occurrence_count_limit
        new.occurrence_spacing = self.occurrence_spacing
        new.start = self.start
        
        return new
    
    
    def copy_with(
        self,
        by_month_days = ...,
        by_months = ...,
        by_nth_weeks_days = ...,
        by_weeks_days = ...,
        by_year_days = ...,
        end = ...,
        frequency = ...,
        occurrence_count_limit = ...,
        occurrence_spacing = ...,
        start = ...,
    ):
        """
        Copies the schedule with the given fields.
        
        Parameters
        ----------
        by_month_days : `None | iterable<int>`, Optional (Keyword only)
            On which days of the month should the event occur at.
        
        by_months : `None | iterable<ScheduleMonth> | iterable<int>`, Optional (Keyword only)
            On which months should the event occur at.
        
        by_nth_weeks_days : `None | iterable<ScheduleNthWeeksDay>`, Optional (Keyword only)
            On which days the event should occur at.
        
        by_weeks_days : `None | iterable<ScheduleWeeksDay> | iterable<int>`, Optional (Keyword only)
            In which days of the week should the event occur at.
        
        by_year_days : `None | iterable<int>`, Optional (Keyword only)
            On which days of the year should the even occur at.
        
        end : `None | DateTime`, Optional (Keyword only)
            When the occurrence should end at.
        
        frequency : `ScheduleFrequency | int`, Optional (Keyword only)
            How often should the event occur.
        
        occurrence_count_limit : `int`, Optional (Keyword only)
            Up to how much times should the event occur.
        
        occurrence_spacing : `int`, Optional (Keyword only)
            The spacing between 2 events.
        
        start : `None | DateTime`, Optional (Keyword only)
            When the occurrence should start at.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # by_month_days
        if by_month_days is ...:
            by_month_days = self.by_month_days
            if (by_month_days is not None):
                by_month_days = (*by_month_days,)
        else:
            by_month_days = validate_by_month_days(by_month_days)
        
        # by_months
        if by_months is ...:
            by_months = self.by_months
            if (by_months is not None):
                by_months = (*by_months,)
        else:
            by_months = validate_by_months(by_months)
        
        # by_nth_weeks_days
        if by_nth_weeks_days is ...:
            by_nth_weeks_days = self.by_nth_weeks_days
            if (by_nth_weeks_days is not None):
                by_nth_weeks_days = (*by_nth_weeks_days,)
        else:
            by_nth_weeks_days = validate_by_nth_weeks_days(by_nth_weeks_days)
        
        # by_weeks_days
        if by_weeks_days is ...:
            by_weeks_days = self.by_weeks_days
            if (by_weeks_days is not None):
                by_weeks_days = (*by_weeks_days,)
        else:
            by_weeks_days = validate_by_weeks_days(by_weeks_days)
        
        # by_year_days
        if by_year_days is ...:
            by_year_days = self.by_year_days
            if (by_year_days is not None):
                by_year_days = (*by_year_days,)
        else:
            by_year_days = validate_by_year_days(by_year_days)
        
        # end
        if end is ...:
            end = self.end
        else:
            end = validate_end(end)
        
        # frequency
        if frequency is ...:
            frequency = self.frequency
        else:
            frequency = validate_frequency(frequency)
        
        # occurrence_count_limit
        if occurrence_count_limit is ...:
            occurrence_count_limit = self.occurrence_count_limit
        else:
            occurrence_count_limit = validate_occurrence_count_limit(occurrence_count_limit)
        
        # occurrence_spacing
        if occurrence_spacing is ...:
            occurrence_spacing = self.occurrence_spacing
        else:
            occurrence_spacing = validate_occurrence_spacing(occurrence_spacing)
        
        # start
        if start is ...:
            start = self.start
        else:
            start = validate_start(start)
        
        # construct
        new = object.__new__(type(self))
        new.by_month_days = by_month_days
        new.by_months = by_months
        new.by_nth_weeks_days = by_nth_weeks_days
        new.by_weeks_days = by_weeks_days
        new.by_year_days = by_year_days
        new.end = end
        new.frequency = frequency
        new.occurrence_count_limit = occurrence_count_limit
        new.occurrence_spacing = occurrence_spacing
        new.start = start
        return new
    
    
    @classmethod
    def create_weekly(cls, day):
        """
        Creates a new schedule that occurs weekly on a specific day.
        
        Parameters
        ----------
        day : ``ScheduleWeeksDay``
            The day to occur at.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return cls(
            by_weeks_days = [day],
            frequency = ScheduleFrequency.weekly,
        )

    
    @classmethod
    def create_bi_weekly(cls, day):
        """
        Creates a new schedule that occurs every other week on a specific day.
        
        Parameters
        ----------
        day : ``ScheduleWeeksDay``
            The day to occur at.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return cls(
            by_weeks_days = [day],
            frequency = ScheduleFrequency.weekly,
            occurrence_spacing = 2,
        )
    
    
    @classmethod
    def create_monthly_nth_weeks_day(cls, nth_week, weeks_day):
        """
        Creates a new schedule that occurs every month on the nth week's same day.
        
        Parameters
        ----------
        nth_week : `int`
            The nth week to re-occur at.
        weeks_day : ``ScheduleWeeksDay``
            The day of the week.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return cls(
            by_nth_weeks_days = [ScheduleNthWeeksDay(nth_week = nth_week, weeks_day = weeks_day)],
            frequency = ScheduleFrequency.monthly,
        )
    
    
    @classmethod
    def create_yearly_month_nth_day(cls, month, nth_day):
        """
        Creates a new schedule that occurs every year on the same month's nth day.
        
        Parameters
        ----------
        month : ``ScheduleMonth``
            The month of the year to re-occur at.
        nth_day : `int`
            The nth day of the month.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return cls(
            by_month_days = [nth_day],
            by_months = [month],
            frequency = ScheduleFrequency.yearly,
        )
